from math import ceil
from typing import Any

from recap.converters.dbapi import DbapiConverter
from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    ProxyType,
    RecapType,
    RecapTypeRegistry,
    StringType,
    UnionType,
)

MAX_FIELD_SIZE = 1073741824

DEFAULT_NAMESPACE = "_root"
"""
Namespace to use when no namespace is specified in the schema.
"""


class PostgresqlConverter(DbapiConverter):
    def __init__(self, namespace: str = DEFAULT_NAMESPACE) -> None:
        self.namespace = namespace
        self.registry = RecapTypeRegistry()

    def _parse_type(self, column_props: dict[str, Any]) -> RecapType:
        column_name = column_props["COLUMN_NAME"]
        data_type = column_props["DATA_TYPE"].lower()
        octet_length = column_props["CHARACTER_OCTET_LENGTH"]
        max_length = column_props["CHARACTER_MAXIMUM_LENGTH"]
        udt_name = (column_props["UDT_NAME"] or "").lower()

        if data_type in ["bigint", "int8", "bigserial", "serial8"]:
            base_type = IntType(bits=64, signed=True)
        elif data_type in ["integer", "int", "int4", "serial", "serial4"]:
            base_type = IntType(bits=32, signed=True)
        elif data_type in ["smallint", "smallserial", "serial2"]:
            base_type = IntType(bits=16, signed=True)
        elif data_type in ["double precision", "float8"]:
            base_type = FloatType(bits=64)
        elif data_type in ["real", "float4"]:
            base_type = FloatType(bits=32)
        elif data_type == "boolean":
            base_type = BoolType()
        elif (
            data_type in ["text", "json", "jsonb"]
            or data_type.startswith("character varying")
            or data_type.startswith("varchar")
        ):
            base_type = StringType(bytes_=octet_length, variable=True)
        elif data_type.startswith("char"):
            base_type = StringType(bytes_=octet_length, variable=False)
        elif data_type == "uuid":
            base_type = StringType(
                logical="build.recap.UUID",
                bytes_=36,
                variable=False,
            )
        elif data_type == "bytea" or data_type.startswith("bit varying"):
            base_type = BytesType(bytes_=MAX_FIELD_SIZE, variable=True)
        elif data_type.startswith("bit"):
            byte_length = ceil(max_length / 8)
            base_type = BytesType(bytes_=byte_length, variable=False)
        elif data_type.startswith("timestamp"):
            dt_precision = column_props["DATETIME_PRECISION"]
            unit = self._get_time_unit([dt_precision]) or "microsecond"
            base_type = IntType(
                bits=64,
                logical="build.recap.Timestamp",
                unit=unit,
            )
        elif data_type.startswith("decimal") or data_type.startswith("numeric"):
            base_type = BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=column_props["NUMERIC_PRECISION"],
                scale=column_props["NUMERIC_SCALE"],
            )
        elif data_type == "array":
            # Remove _ for standard PG types like _int4
            nested_data_type = udt_name.lstrip("_")
            # Recurse to get the array value type (the int4 in int4[])
            # Postgres arrays ignore value type octet lengths for varchars, bit
            # lengths, etc. Thus, we only set DATA_TYPE here. Sigh.
            value_type = self._parse_type(
                {
                    "COLUMN_NAME": None,
                    "DATA_TYPE": nested_data_type,
                    # Default strings, bits, etc. to the max field size since
                    # information_schema doesn't contain lengths for array
                    # types.
                    # TODO Technically, we could consult pg_attribute and
                    # pg_type for this information, but that's not implemented
                    # right now.
                    "CHARACTER_OCTET_LENGTH": MAX_FIELD_SIZE,
                    # * 8 because bit columns use bits not bytes.
                    "CHARACTER_MAXIMUM_LENGTH": MAX_FIELD_SIZE * 8,
                    "UDT_NAME": None,
                }
            )
            column_name_without_periods = column_name.replace(".", "_")
            base_type_alias = f"{self.namespace}.{column_name_without_periods}"
            # Construct a self-referencing list comprised of the array's value
            # type and a proxy to the list itself. This allows arrays to be an
            # arbitrary number of dimensions, which is how PostgreSQL treats
            # lists. See https://github.com/recap-build/recap/issues/264 for
            # more details.
            base_type = ListType(
                alias=base_type_alias,
                values=UnionType(
                    types=[
                        value_type,
                        ProxyType(
                            alias=base_type_alias,
                            registry=self.registry,
                        ),
                    ],
                ),
            )
            self.registry.register_alias(base_type)
        else:
            raise ValueError(f"Unknown data type: {data_type}")

        return base_type



import pytest

from recap.clients.postgresql import PostgresqlClient


@pytest.mark.parametrize(
    "method, url_args, expected_result",
    [
        # Test "ls" method with only URL (no catalog or schema)
        (
            "ls",
            {"scheme": "postgres", "netloc": "postgres-url"},
            ("postgres://postgres-url", [None, None]),
        ),
        # Test "ls" method with URL and catalog
        (
            "ls",
            {"scheme": "postgres", "netloc": "postgres-url", "paths": ["cat1"]},
            ("postgres://postgres-url/cat1", ["cat1", None]),
        ),
        # Test "ls" method with URL, catalog, and schema
        (
            "ls",
            {"scheme": "postgres", "netloc": "postgres-url", "paths": ["cat1", "sch1"]},
            ("postgres://postgres-url/cat1", ["cat1", "sch1"]),
        ),
        # Test "schema" method with URL, catalog, schema, and table
        (
            "schema",
            {
                "scheme": "postgres",
                "netloc": "postgres-url",
                "paths": ["cat1", "sch1", "tbl1"],
            },
            ("postgres://postgres-url/cat1", ["cat1", "sch1", "tbl1"]),
        ),
        # Test invalid method
        (
            "invalid_method",
            {"scheme": "postgres", "netloc": "postgres-url"},
            pytest.raises(ValueError, match="Invalid method"),
        ),
        # Test "schema" method with insufficient paths (only catalog and schema, no table)
        (
            "schema",
            {"scheme": "postgres", "netloc": "postgres-url", "paths": ["cat1", "sch1"]},
            pytest.raises(ValueError, match="Invalid method"),
        ),
        # Test "ls" method with all paths (catalog, schema, and table)
        (
            "ls",
            {
                "scheme": "postgres",
                "netloc": "postgres-url",
                "paths": ["cat1", "sch1", "tbl1"],
            },
            pytest.raises(ValueError, match="Invalid method"),
        ),
    ],
)
def test_parse_method(method, url_args, expected_result):
    if isinstance(expected_result, tuple):
        result = PostgresqlClient.parse(method, **url_args)
        assert result == expected_result
    else:
        with expected_result:
            PostgresqlClient.parse(method, **url_args)



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator

from recap.types import StructType


class FilesystemClient:
    def __init__(self, scheme: str):
        self.scheme = scheme

    @staticmethod
    @contextmanager
    def create(scheme: str, **_) -> Generator[FilesystemClient, None, None]:
        yield FilesystemClient(scheme=scheme)

    @staticmethod
    def parse(_: str, **url_args) -> tuple[str, list[Any]]:
        from urllib.parse import urlsplit

        split_url = urlsplit(url_args["url"])
        root_url = f"{split_url.scheme or 'file'}://"
        path = "/"

        if split_url.netloc:
            path += split_url.netloc + "/"
        if split_url.path:
            path += split_url.path[1:] + "/"

        return (root_url, [path])

    def ls(self, path: str = "/") -> list[str]:
        if self.scheme == "file":
            return [p.name for p in Path(path).iterdir()]
        else:
            import fsspec

            fs = fsspec.filesystem(self.scheme)
            return [Path(p).name for p in fs.ls(path)]

    def schema(self, path: str = "/") -> StructType:
        suffix = Path(path).suffix.lower().strip(".")

        match suffix:
            case "avsc" | "avro" | "avdl" | "avpr":
                from recap.converters.avro import AvroConverter

                return AvroConverter().to_recap(self._read_file(path))
            case "json" | "jsonschema" | "jsons":
                from recap.converters.json_schema import JSONSchemaConverter

                return JSONSchemaConverter().to_recap(self._read_file(path))
            case "proto" | "protobuf" | "protos" | "proto3" | "proto2":
                from recap.converters.protobuf import ProtobufConverter

                return ProtobufConverter().to_recap(self._read_file(path))
            case _:
                raise ValueError(f"Unsupported file type {suffix}")

    def _read_file(self, path: str) -> str:
        if self.scheme == "file":
            return Path(path).read_text()
        else:
            import fsspec

            fs = fsspec.filesystem(self.scheme)
            return fs.read_text(path)



from unittest.mock import patch

import pytest

from recap.clients.fs import FilesystemClient
from recap.types import IntType, StringType, StructType


def test_avro_schema():
    # Mock the `_read_file` method to return a sample Avro schema string.
    mock_avro_schema_str = """
    {
      "type": "record",
      "name": "User",
      "fields": [
        {"name": "name", "type": "string"},
        {"name": "age", "type": "int"}
      ]
    }
    """

    # Create an instance of FilesystemClient with scheme "file"
    client = FilesystemClient(scheme="file")

    # Mock the `_read_file` method
    with patch.object(client, "_read_file", return_value=mock_avro_schema_str):
        result = client.schema("dummy_user.avsc")

    # Check that the schema was converted correctly
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert result.fields[0].extra_attrs["name"] == "name"
    assert isinstance(result.fields[0], StringType)
    assert result.fields[1].extra_attrs["name"] == "age"
    assert isinstance(result.fields[1], IntType)


def test_json_schema():
    # Mock the `_read_file` method to return a sample JSON schema string.
    mock_json_schema_str = """
    {
      "$schema": "http://json-schema.org/draft-07/schema#",
      "type": "object",
      "title": "User",
      "properties": {
        "name": {
          "type": "string"
        },
        "age": {
          "type": "integer"
        }
      },
      "required": ["name", "age"]
    }
    """

    # Create an instance of FilesystemClient with scheme "file"
    client = FilesystemClient(scheme="file")

    # Mock the `_read_file` method
    with patch.object(client, "_read_file", return_value=mock_json_schema_str):
        result = client.schema("dummy_user.json")

    # Check that the schema was converted correctly
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert result.fields[0].extra_attrs["name"] == "name"
    assert isinstance(result.fields[0], StringType)
    assert result.fields[1].extra_attrs["name"] == "age"
    assert isinstance(result.fields[1], IntType)


def test_proto3_schema():
    # Mock the `_read_file` method to return a sample proto3 schema string.
    mock_proto3_schema_str = """
    syntax = "proto3";

    message User {
        required string name = 1;
        required int32 age = 2;
    }
    """

    # Create an instance of FilesystemClient with scheme "file"
    client = FilesystemClient(scheme="file")

    # Mock the `_read_file` method
    with patch.object(client, "_read_file", return_value=mock_proto3_schema_str):
        result = client.schema("dummy_user.proto")

    # Check that the schema was converted correctly
    assert isinstance(result, StructType)
    assert len(result.fields) == 2
    assert result.fields[0].extra_attrs["name"] == "name"
    assert isinstance(result.fields[0], StringType)
    assert result.fields[1].extra_attrs["name"] == "age"
    assert isinstance(result.fields[1], IntType)


@pytest.mark.parametrize(
    "url_input,expected_output",
    [
        # Test with a standard http URL
        (
            "http://example.com/path/to/resource",
            ("http://", ["/example.com/path/to/resource/"]),
        ),
        # Test with an https URL and a deeper path
        (
            "https://example.com/path/to/resource/file.ext",
            ("https://", ["/example.com/path/to/resource/file.ext/"]),
        ),
        # Test with a file scheme
        ("file:///path/to/local/file", ("file://", ["/path/to/local/file/"])),
        # Test with no scheme
        ("/path/to/local/file", ("file://", ["/path/to/local/file/"])),
        # Add other test cases as needed
    ],
)
def test_filesystemclient_parse(url_input, expected_output):
    # Given a URL, parse it using the static method
    result = FilesystemClient.parse("", url=url_input)

    # Check that the result matches the expected output
    assert result == expected_output



import mysql.connector

from recap.clients import create_client
from recap.clients.mysql import MysqlClient
from recap.types import (
    BytesType,
    FloatType,
    IntType,
    NullType,
    StringType,
    StructType,
    UnionType,
)


class TestMySqlClient:
    @classmethod
    def setup_class(cls):
        # Connect to the MySQL database
        cls.connection = mysql.connector.connect(
            database="testdb",
            user="mysql",
            password="password",
            host="localhost",
            port=3306,
        )

        # Create tables
        cursor = cls.connection.cursor()
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS test_types (
                test_bigint BIGINT,
                test_integer INTEGER,
                test_smallint SMALLINT,
                test_float FLOAT,
                test_double DOUBLE PRECISION,
                test_real REAL,
                test_boolean BOOLEAN,
                test_text TEXT,
                test_char CHAR(10),
                test_blob BLOB,
                test_mediumblob MEDIUMBLOB,
                test_longblob LONGBLOB,
                test_mediumtext MEDIUMTEXT,
                test_longtext LONGTEXT,
                test_tinyblob TINYBLOB,
                test_tinytext TINYTEXT,
                test_varchar VARCHAR(255),
                test_varbinary VARBINARY(255),
                test_binary BINARY(255),
                test_bit BIT(10),
                test_timestamp TIMESTAMP,
                test_timestamp_millis TIMESTAMP(3),
                test_decimal DECIMAL(10,2),
                test_numeric NUMERIC(11,3),
                test_not_null INTEGER NOT NULL,
                test_not_null_default INTEGER NOT NULL DEFAULT 1,
                test_default INTEGER DEFAULT 2
            );
        """
        )
        cls.connection.commit()

    @classmethod
    def teardown_class(cls):
        # Delete tables
        cursor = cls.connection.cursor()
        cursor.execute("DROP TABLE IF EXISTS test_types;")
        cls.connection.commit()

        # Close the connection
        cls.connection.close()

    def test_struct_method(self):
        # Initiate the MySQL client class
        client = MysqlClient(self.connection)  # type: ignore

        # Test 'test_types' table. MySQL catalog is always 'def'.
        test_types_struct = client.schema("testdb", "test_types")

        # Define the expected output for 'test_types' table
        expected_fields = [
            UnionType(
                default=None,
                name="test_bigint",
                types=[NullType(), IntType(bits=64, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_integer",
                types=[NullType(), IntType(bits=32, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_smallint",
                types=[NullType(), IntType(bits=16, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_float",
                types=[NullType(), FloatType(bits=32)],
            ),
            UnionType(
                default=None,
                name="test_double",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="test_real",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="test_boolean",
                types=[NullType(), IntType(bits=8, signed=True)],
            ),
            UnionType(
                default=None,
                name="test_text",
                types=[NullType(), StringType(bytes_=65_535, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_char",
                # 40 = max of 4 bytes in a UTF-8 encoded unicode character * 10 chars
                types=[NullType(), StringType(bytes_=40, variable=False)],
            ),
            UnionType(
                default=None,
                name="test_blob",
                types=[NullType(), BytesType(bytes_=65_535, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_mediumblob",
                types=[NullType(), BytesType(bytes_=16_777_215, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_longblob",
                types=[NullType(), BytesType(bytes_=4_294_967_295, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_mediumtext",
                types=[NullType(), StringType(bytes_=16_777_215, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_longtext",
                types=[NullType(), StringType(bytes_=4_294_967_295, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_tinyblob",
                types=[NullType(), BytesType(bytes_=255, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_tinytext",
                types=[NullType(), StringType(bytes_=255, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_varchar",
                types=[
                    NullType(),
                    StringType(bytes_=1020, variable=True),
                ],  # max 4 bytes per char * 255
            ),
            UnionType(
                default=None,
                name="test_varbinary",
                types=[NullType(), BytesType(bytes_=255, variable=True)],
            ),
            UnionType(
                default=None,
                name="test_binary",
                types=[NullType(), BytesType(bytes_=255, variable=False)],
            ),
            UnionType(
                default=None,
                name="test_bit",
                types=[NullType(), BytesType(bytes_=8, variable=False)],
            ),
            UnionType(
                default=None,
                name="test_timestamp",
                types=[
                    NullType(),
                    IntType(
                        bits=64,
                        logical="build.recap.Timestamp",
                        unit="second",
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_timestamp_millis",
                types=[
                    NullType(),
                    IntType(
                        bits=64,
                        logical="build.recap.Timestamp",
                        unit="millisecond",
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_decimal",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=32,
                        variable=False,
                        precision=10,
                        scale=2,
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="test_numeric",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=32,
                        variable=False,
                        precision=11,
                        scale=3,
                    ),
                ],
            ),
            IntType(bits=32, signed=True, name="test_not_null"),
            IntType(bits=32, signed=True, name="test_not_null_default", default=b"1"),
            UnionType(
                default=b"2",
                name="test_default",
                types=[NullType(), IntType(bits=32, signed=True)],
            ),
        ]

        # Going field by field to make debugging easier when test fails
        for field, expected_field in zip(test_types_struct.fields, expected_fields):
            assert field == expected_field

        assert test_types_struct == StructType(fields=expected_fields)

    def test_create_client(self):
        mysql_url = "mysql://mysql:password@localhost:3306/testdb"

        with create_client(mysql_url) as client:
            assert client.ls() == ["information_schema", "performance_schema", "testdb"]
            assert client.ls("testdb") == ["test_types"]



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Generator

from confluent_kafka.schema_registry import SchemaRegistryClient

from recap.types import StructType

ALLOWED_ATTRS = {
    "url",
    "ssl.ca.location",
    "ssl.key.location",
    "ssl.certificate.location",
    "basic.auth.user.info",
}


class ConfluentRegistryClient:
    def __init__(self, registry: SchemaRegistryClient):
        self.registry = registry

    @staticmethod
    @contextmanager
    def create(**url_args) -> Generator[ConfluentRegistryClient, None, None]:
        # Filter out kwargs that are not allowed by the SchemaRegistryClient
        url_args = {k: v for k, v in url_args.items() if k in ALLOWED_ATTRS}

        with SchemaRegistryClient(url_args) as registry:
            yield ConfluentRegistryClient(registry)

    @staticmethod
    def parse(method: str, **url_args) -> tuple[str, list[Any]]:
        from urllib.parse import urlunparse

        match method:
            case "ls":
                url_with_clean_scheme = (
                    url_args["url"]
                    .replace("http+csr://", "http://")
                    .replace("https+csr://", "https://")
                )
                return (url_with_clean_scheme, [])
            case "schema":
                subject = url_args["paths"].pop(-1)
                connection_url = urlunparse(
                    [
                        url_args.get("dialect") or url_args.get("scheme"),
                        url_args.get("netloc"),
                        "/".join(url_args.get("paths", [])),
                        url_args.get("params"),
                        url_args.get("query"),
                        url_args.get("fragment"),
                    ]
                )
                return (connection_url, [subject])
            case _:
                raise ValueError("Invalid method")

    def ls(self) -> list[str]:
        return self.registry.get_subjects()

    def schema(self, subject: str) -> StructType:
        has_kv = subject.endswith("-key") or subject.endswith("-value")
        subject = subject if has_kv else f"{subject}-value"
        registered_schema = self.registry.get_latest_version(subject)
        schema_str = registered_schema.schema.schema_str
        match registered_schema.schema.schema_type:
            case "AVRO":
                from recap.converters.avro import AvroConverter

                return AvroConverter().to_recap(schema_str)
            case "JSON":
                from recap.converters.json_schema import JSONSchemaConverter

                return JSONSchemaConverter().to_recap(schema_str)
            case "PROTOBUF":
                from recap.converters.protobuf import ProtobufConverter

                return ProtobufConverter().to_recap(schema_str)
            case _:
                raise ValueError(
                    f"Unsupported schema type {registered_schema.schema.schema_type}"
                )



import os
from pathlib import Path
from typing import Any
from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit

from dotenv import load_dotenv
from pydantic import AnyUrl, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

load_dotenv()

HOME_PATH = Path(os.environ.get("RECAP_HOME") or os.path.expanduser("~/.recap"))
DEFAULT_REGISTRY_STORAGE_PATH = Path(HOME_PATH, "schemas")
SECRETS_PATH = Path(dir) if (dir := os.environ.get("RECAP_SECRETS")) else None


def mkdirs():
    HOME_PATH.mkdir(exist_ok=True)

    if SECRETS_PATH:
        SECRETS_PATH.mkdir(exist_ok=True)


mkdirs()


class RecapSettings(BaseSettings):
    urls: list[AnyUrl] = Field(default_factory=list)
    registry_storage_url: AnyUrl = Field(default=DEFAULT_REGISTRY_STORAGE_PATH.as_uri())
    registry_storage_url_args: dict[str, Any] = Field(default_factory=dict)
    model_config = SettingsConfigDict(
        env_file_encoding="utf-8",
        env_prefix="recap_",
        env_nested_delimiter="__",
        secrets_dir=str(SECRETS_PATH) if SECRETS_PATH else None,
    )

    @property
    def safe_urls(self) -> list[str]:
        """
        Return a list of URLs that have the username and password removed.
        """

        safe_urls_list = []

        for url in self.urls:
            split_url = urlsplit(str(url))
            netloc = split_url.netloc

            if split_url.username or split_url.password:
                if split_url.hostname is not None:
                    netloc = split_url.hostname

                if split_url.port is not None:
                    netloc += f":{split_url.port}"

            sanitized_url = urlunsplit(
                (
                    split_url.scheme,
                    netloc,
                    split_url.path.strip("/"),
                    split_url.query,
                    split_url.fragment,
                )
            )

            safe_urls_list.append(sanitized_url)

        return safe_urls_list

    def unsafe_url(self, url: str, strict: bool = True) -> str:
        """
        Merge the a URL's user, pass, path, query, and fragment in settings
        into the input URL and return it. Merging occurs if:

        - The input URL's scheme, hostname, and port match a URL in settings.
        - The input URL's path starts with the settings URL's path.

        :param url: URL to merge
        :param strict: If True, raise an error if the URL is not configured in settings.
        :return: URL with user, pass, path, query, and fragment merged from settings.
        """

        url_split = urlsplit(url)
        url_path = Path(url_split.path or "/").as_posix()

        for unsafe_url in self.urls:
            unsafe_url_split = urlsplit(unsafe_url.unicode_string())

            if (
                unsafe_url_split.scheme == url_split.scheme
                and unsafe_url_split.hostname == url_split.hostname
                and unsafe_url_split.port == url_split.port
                and url_path.startswith(Path(unsafe_url_split.path).as_posix())
            ):
                netloc = unsafe_url_split.netloc

                # Merge query strings.
                # NOTE: This doesn't work for URLs that have multiple values
                # for the same key spanning both the unsafe and input URLs.
                unsafe_qs = parse_qs(unsafe_url_split.query)
                url_qs = parse_qs(url_split.query)
                merged_qs = unsafe_qs | url_qs
                query = urlencode(merged_qs, doseq=True)

                merged_url = urlunsplit(
                    (
                        url_split.scheme,
                        netloc,
                        url_path.strip("/"),
                        query,
                        url_split.fragment or unsafe_url_split.fragment,
                    )
                )

                # Unsplit returns a URL with a trailing colon if the URL only
                # has a scheme. This looks weird, so include trailing double
                # slash (e.g. bigquery: to bigquery://).
                if merged_url == f"{url_split.scheme}:":
                    merged_url += "//"

                return merged_url

        if strict:
            raise ValueError(f"URL must be configured in Recap settings: {url}")

        return url



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Generator

from recap.clients.dbapi import Connection, DbapiClient
from recap.converters.mysql import MysqlConverter
from recap.types import StructType

MYSQL_CATALOG_NAME = "def"


class MysqlClient(DbapiClient):
    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, MysqlConverter())

    @staticmethod
    @contextmanager
    def create(
        paths: list[str] | None = None,
        **url_args,
    ) -> Generator[MysqlClient, None, None]:
        import mysql.connector
        from mysql.connector.constants import DEFAULT_CONFIGURATION

        if paths:
            url_args["database"] = paths[0]

        # Only include kwargs that are valid for the MySQL connector
        kwargs = {k: v for k, v in url_args.items() if k in DEFAULT_CONFIGURATION}

        with mysql.connector.connect(**kwargs) as client:
            yield MysqlClient(client)  # type: ignore

    @staticmethod
    def parse(method: str, **url_args) -> tuple[str, list[Any]]:
        from urllib.parse import urlunparse

        paths = url_args.get("paths", [])[::-1]
        schema = paths.pop() if paths else None
        table = paths.pop() if paths else None
        connection_url = urlunparse(
            [
                url_args.get("scheme"),
                url_args.get("netloc"),
                schema or "",
                url_args.get("params"),
                url_args.get("query"),
                url_args.get("fragment"),
            ]
        )

        match method:
            case "ls":
                return (connection_url, [schema])
            case "schema" if schema and table:
                return (connection_url, [schema, table])
            case _:
                raise ValueError("Invalid method")

    def ls(self, schema: str | None = None) -> list[str]:
        match schema:
            case None:
                return self.ls_schemas(MYSQL_CATALOG_NAME)
            case str(schema):
                return self.ls_tables(MYSQL_CATALOG_NAME, schema)
            case _:
                raise ValueError("Invalid arguments")

    def schema(self, schema: str, table: str) -> StructType:
        return super().schema(MYSQL_CATALOG_NAME, schema, table)



import pytest

from recap.converters.postgresql import MAX_FIELD_SIZE, PostgresqlConverter
from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    ProxyType,
    StringType,
    UnionType,
)


@pytest.mark.parametrize(
    "column_props,expected",
    [
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "bigint",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            IntType(bits=64, signed=True),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "int",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            IntType(bits=32, signed=True),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "smallint",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            IntType(bits=16, signed=True),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "double precision",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            FloatType(bits=64),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "real",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            FloatType(bits=32),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "boolean",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            BoolType(),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "text",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": 65536,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            StringType(bytes_=65536, variable=True),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "character varying",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": 255,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            StringType(bytes_=255, variable=True),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "char",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": 255,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            StringType(bytes_=255, variable=False),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "bytea",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            BytesType(bytes_=MAX_FIELD_SIZE),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "bit",
                "CHARACTER_MAXIMUM_LENGTH": 1,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            BytesType(bytes_=1, variable=False),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "bit",
                "CHARACTER_MAXIMUM_LENGTH": 17,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
            },
            BytesType(bytes_=3, variable=False),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "timestamp",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
                "DATETIME_PRECISION": 3,
            },
            IntType(bits=64, logical="build.recap.Timestamp", unit="millisecond"),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "timestamp",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": None,
                "NUMERIC_SCALE": None,
                "UDT_NAME": None,
                "DATETIME_PRECISION": 3,
            },
            IntType(bits=64, logical="build.recap.Timestamp", unit="millisecond"),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "decimal",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 10,
                "NUMERIC_SCALE": 2,
                "UDT_NAME": None,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=10,
                scale=2,
            ),
        ),
        (
            {
                "COLUMN_NAME": "test_column",
                "DATA_TYPE": "numeric",
                "CHARACTER_MAXIMUM_LENGTH": None,
                "CHARACTER_OCTET_LENGTH": None,
                "NUMERIC_PRECISION": 5,
                "NUMERIC_SCALE": 0,
                "UDT_NAME": None,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=5,
                scale=0,
            ),
        ),
    ],
)
def test_postgresql_converter(column_props, expected):
    result = PostgresqlConverter()._parse_type(column_props)
    assert result == expected


def test_postgresql_converter_array():
    converter = PostgresqlConverter()
    column_props = {
        "COLUMN_NAME": "test_column",
        "DATA_TYPE": "array",
        "CHARACTER_MAXIMUM_LENGTH": None,
        "CHARACTER_OCTET_LENGTH": None,
        "NUMERIC_PRECISION": 5,
        "NUMERIC_SCALE": 0,
        "UDT_NAME": "_int4",
    }
    expected = ListType(
        alias="_root.test_column",
        values=UnionType(
            types=[
                IntType(bits=32, signed=True),
                ProxyType(
                    alias="_root.test_column",
                    registry=converter.registry,
                ),
            ],
        ),
    )
    result = converter._parse_type(column_props)
    assert result == expected



import os
import tempfile
from json import loads
from unittest.mock import patch

from typer.testing import CliRunner

from recap.cli import app
from recap.types import IntType, StructType, to_dict

runner = CliRunner()


class TestCli:
    @classmethod
    def setup_class(cls):
        # Create a temporary .env file for the test
        cls.temp_file = tempfile.NamedTemporaryFile(delete=False)
        cls.temp_file_name = cls.temp_file.name

    @classmethod
    def teardown_class(cls):
        # Delete the temporary .env file
        os.unlink(cls.temp_file_name)

    @patch("recap.commands.ls")
    def test_ls_root(self, mock_ls):
        mock_ls.return_value = ["foo", "bar"]
        result = runner.invoke(app, ["ls"])
        assert result.exit_code == 0
        assert loads(result.stdout) == mock_ls.return_value

    @patch("recap.commands.ls")
    def test_ls_subpath(self, mock_ls):
        mock_ls.return_value = ["foo", "bar"]
        result = runner.invoke(app, ["ls", "/foo/bar"])
        assert result.exit_code == 0
        assert loads(result.stdout) == mock_ls.return_value

    @patch("recap.commands.schema")
    def test_schema(self, mock_schema):
        mock_schema.return_value = to_dict(StructType([IntType(bits=32)]))
        result = runner.invoke(app, ["schema", "foo"])
        assert result.exit_code == 0
        assert loads(result.stdout) == {"type": "struct", "fields": ["int32"]}



from unittest.mock import patch

from fastapi.testclient import TestClient

from recap.server.gateway import router

client = TestClient(router)


@patch("recap.commands.ls")
def test_ls_root(mock_ls):
    mock_ls.return_value = ["foo", "bar"]
    response = client.get("/gateway/ls")
    assert response.status_code == 200
    assert response.json() == mock_ls.return_value


@patch("recap.commands.ls")
def test_ls_subpath(mock_ls):
    mock_ls.return_value = ["foo", "bar"]
    response = client.get("/gateway/ls/bar")
    assert response.status_code == 200
    assert response.json() == mock_ls.return_value


@patch("recap.commands.schema")
def test_schema(mock_schema):
    mock_schema.return_value = {"type": "struct", "fields": ["int32"]}
    response = client.get("/gateway/schema/foo")
    expected = {"type": "struct", "fields": ["int32"]}
    assert response.status_code == 200
    assert response.json() == expected



from fastapi import APIRouter, HTTPException, Request

from recap import commands

router = APIRouter(prefix="/gateway")


@router.get("/ls/{url:path}")
async def ls(url: str | None = None) -> list[str]:
    """
    List the children of a URL.
    """

    children = commands.ls(url)
    if children is not None:
        return children
    raise HTTPException(status_code=404, detail="URL not found")


@router.get("/schema/{url:path}")
async def schema(url: str, request: Request):
    """
    Get the schema of a URL.
    """

    content_type = request.headers.get("content-type") or "application/x-recap+json"
    if format := commands.FORMAT_MAP.get(content_type):
        return commands.schema(url, format)
    else:
        raise HTTPException(
            status_code=415,
            detail=f"Unsupported content type: {content_type}",
        )



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Generator

from pymetastore.metastore import HMS, HTable
from pymetastore.stats import (
    BinaryTypeStats,
    BooleanTypeStats,
    DateTypeStats,
    DecimalTypeStats,
    DoubleTypeStats,
    LongTypeStats,
    StringTypeStats,
    TypeStats,
)

from recap.converters.hive_metastore import HiveMetastoreConverter
from recap.types import RecapType, StructType


class HiveMetastoreClient:
    def __init__(self, client: HMS):
        self.client = client

    @staticmethod
    @contextmanager
    def create(host: str, port: int, **_) -> Generator[HiveMetastoreClient, None, None]:
        with HMS.create(host, port) as client:
            yield HiveMetastoreClient(client)

    @staticmethod
    def parse(
        method: str,
        url: str,
        paths: list[str],
        include_stats: str | None = None,
        **url_args,
    ) -> tuple[str, list[Any]]:
        clean_url = f"thrift://{url_args['host']}"
        if url_args["port"]:
            clean_url += f":{url_args['port']}"
        match method:
            case "ls":
                return (clean_url, paths)
            case "schema" if len(paths) >= 2:
                return (clean_url, paths + [include_stats is not None])
            case _:
                raise ValueError("Invalid method")

    def ls(self, database: str | None = None, table: str | None = None) -> list[str]:
        match (database, table):
            case (None, None):
                return self.ls_databases()
            case (str(database), None):
                return self.ls_tables(database)
            case (str(database), str(table)):
                return [table]
            case _:
                raise ValueError("Invalid arguments")

    def ls_databases(self) -> list[str]:
        return self.client.list_databases()

    def ls_tables(self, database: str) -> list[str]:
        return self.client.list_tables(database)

    def schema(
        self,
        database: str,
        table: str,
        include_stats: bool = False,
    ) -> StructType:
        htable = self.client.get_table(database, table)
        struct = HiveMetastoreConverter().to_recap(htable)
        if include_stats:
            self._load_stats(htable, struct.fields)
        return struct

    def _load_stats(self, table: HTable, fields: list[RecapType]) -> None:
        fields_dict = {f.extra_attrs["name"]: f for f in fields}
        table_stats = self.client.get_table_stats(table, [])
        table_stats = [cs for cs in table_stats if cs.stats is not None]
        for col_stats in table_stats:
            recap_type = fields_dict[col_stats.columnName]
            match col_stats.stats:
                case LongTypeStats() | DoubleTypeStats():
                    recap_type.extra_attrs |= {
                        "low": col_stats.stats.lowValue,
                        "high": col_stats.stats.highValue,
                        "cardinality": col_stats.stats.cardinality,
                    }
                case DateTypeStats():
                    stats: dict[str, Any] = {
                        "cardinality": col_stats.stats.cardinality,
                    }
                    if col_stats.stats.lowValue is not None:
                        stats["low"] = col_stats.stats.lowValue.daysSinceEpoch
                    if col_stats.stats.highValue is not None:
                        stats["high"] = col_stats.stats.highValue.daysSinceEpoch
                    recap_type.extra_attrs |= stats
                case StringTypeStats():
                    recap_type.extra_attrs |= {
                        "average_length": col_stats.stats.avgColLen,
                        "max_length": col_stats.stats.maxColLen,
                        "cardinality": col_stats.stats.cardinality,
                    }
                case BooleanTypeStats():
                    recap_type.extra_attrs |= {
                        "true_count": col_stats.stats.numTrues,
                        "false_count": col_stats.stats.numFalses,
                    }
                case BinaryTypeStats():
                    recap_type.extra_attrs |= {
                        "average_length": col_stats.stats.avgColLen,
                        "max_length": col_stats.stats.maxColLen,
                    }
                case DecimalTypeStats():
                    stats: dict[str, Any] = {
                        "cardinality": col_stats.stats.cardinality,
                    }
                    if col_stats.stats.lowValue is not None:
                        stats["low"] = col_stats.stats.lowValue
                    if col_stats.stats.highValue is not None:
                        stats["high"] = col_stats.stats.highValue
                    recap_type.extra_attrs |= stats

            if isinstance(col_stats.stats, TypeStats):
                recap_type.extra_attrs["null_count"] = col_stats.stats.numNulls



import pytest

from recap.settings import RecapSettings


@pytest.mark.parametrize(
    "input_urls, expected_output",
    [
        # Test with no credentials
        (["http://example.com"], ["http://example.com"]),
        # Test with a username
        (["http://user@example.com"], ["http://example.com"]),
        # Test with a username and password
        (["http://user:pass@example.com"], ["http://example.com"]),
        # Test with a port, username, and password
        (["http://user:pass@example.com:8080"], ["http://example.com:8080"]),
        # Test with multiple URLs
        (
            ["http://user:pass@example.com", "https://secure.site"],
            ["http://example.com", "https://secure.site"],
        ),
        # Test with similar schemes but different credentials
        (
            ["http://user1:pass1@example.com", "http://user2:pass2@example.com"],
            ["http://example.com", "http://example.com"],
        ),
        # Tests with queries
        (["http://example.com?query=value"], ["http://example.com?query=value"]),
        (
            ["http://user:pass@example.com?query=value"],
            ["http://example.com?query=value"],
        ),
        # Tests with fragments
        (["http://example.com#fragment"], ["http://example.com#fragment"]),
        (["http://user:pass@example.com#fragment"], ["http://example.com#fragment"]),
        # Tests with queries and fragments
        (
            ["http://example.com?query=value#fragment"],
            ["http://example.com?query=value#fragment"],
        ),
        (
            ["http://user:pass@example.com?query=value#fragment"],
            ["http://example.com?query=value#fragment"],
        ),
        # Multiple URLs with mixed properties
        (
            ["http://example.com?query=value", "http://user:pass@example.com#fragment"],
            ["http://example.com?query=value", "http://example.com#fragment"],
        ),
    ],
)
def test_safe_urls(input_urls, expected_output):
    settings = RecapSettings(urls=input_urls)
    assert settings.safe_urls == expected_output


@pytest.mark.parametrize(
    "stored_urls, merging_url, expected_output, strict",
    [
        # Basic test with scheme, host, and port match
        (
            ["http://user:pass@example.com"],
            "http://example.com",
            "http://user:pass@example.com",
            True,
        ),
        # Test with a different path in the merging URL
        (
            ["http://user:pass@example.com"],
            "http://example.com/path",
            "http://user:pass@example.com/path",
            True,
        ),
        # Test with a query in the merging URL
        (
            ["http://user:pass@example.com"],
            "http://example.com?query=value",
            "http://user:pass@example.com?query=value",
            True,
        ),
        # Test with a fragment in the merging URL
        (
            ["http://user:pass@example.com"],
            "http://example.com#fragment",
            "http://user:pass@example.com#fragment",
            True,
        ),
        # Test with both a query and a fragment in the merging URL
        (
            ["http://user:pass@example.com"],
            "http://example.com?query=value#fragment",
            "http://user:pass@example.com?query=value#fragment",
            True,
        ),
        # Merging query parameters from both stored and merging URL
        (
            ["http://user:pass@example.com?stored=query"],
            "http://example.com?query=value",
            "http://user:pass@example.com?stored=query&query=value",
            True,
        ),
        # Merging fragments from both stored and merging URL (prefer merging URL's fragment)
        (
            ["http://user:pass@example.com#stored_fragment"],
            "http://example.com#fragment",
            "http://user:pass@example.com#fragment",
            True,
        ),
        # Basic merging of paths
        (
            ["http://user:pass@example.com/path1"],
            "http://example.com/path1/path2",
            "http://user:pass@example.com/path1/path2",
            True,
        ),
        # Merging path and query
        (
            ["http://user:pass@example.com/path1?query1=value1"],
            "http://example.com/path1/path2?query2=value2",
            "http://user:pass@example.com/path1/path2?query1=value1&query2=value2",
            True,
        ),
        # Merging path and fragment
        (
            ["http://user:pass@example.com/path1#fragment1"],
            "http://example.com/path1/path2#fragment2",
            "http://user:pass@example.com/path1/path2#fragment2",
            True,
        ),
        # Merging path, query, and fragment
        (
            ["http://user:pass@example.com/path1?query1=value1#fragment1"],
            "http://example.com/path1/path2?query2=value2#fragment2",
            "http://user:pass@example.com/path1/path2?query1=value1&query2=value2#fragment2",
            True,
        ),
        # When both URLs have the same paths but different queries or fragments
        (
            ["http://user:pass@example.com/path"],
            "http://example.com/path?query=value",
            "http://user:pass@example.com/path?query=value",
            True,
        ),
        # Test unstrict mode
        (
            ["http://some-other.com"],
            "http://example.com/path",
            "http://example.com/path",
            False,
        ),
    ],
)
def test_unsafe_url(stored_urls, merging_url, expected_output, strict):
    settings = RecapSettings(urls=stored_urls)
    assert settings.unsafe_url(merging_url, strict) == expected_output


@pytest.mark.parametrize(
    "stored_urls, merging_url",
    [
        # Different paths
        (["http://example.com/path1"], "http://example.com/path2"),
        # Shorter path
        (["http://example.com/path1/path2"], "http://example.com/path1"),
        # Different ports
        (["http://example.com:8080"], "http://example.com:9090"),
        # Different schemes
        (["http://example.com"], "https://example.com"),
        # Different hostnames
        (["http://example1.com"], "http://example2.com"),
    ],
)
def test_unsafe_url_exceptions(stored_urls, merging_url):
    settings = RecapSettings(urls=stored_urls)

    with pytest.raises(
        ValueError, match=f"URL must be configured in Recap settings: {merging_url}"
    ):
        settings.unsafe_url(merging_url)



# pylint: disable=missing-docstring

import re

import pytest

from recap.types import (
    ALIAS_REGEX,
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    RecapType,
    RecapTypeRegistry,
    StringType,
    StructType,
    UnionType,
    from_dict,
    to_dict,
)


def test_null_type():
    test_dict = {"type": "null"}
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, NullType)
    assert recap_type.type_ == "null"


def test_bool_type():
    test_dict = {"type": "bool"}
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, BoolType)
    assert recap_type.type_ == "bool"


def test_int_type():
    test_dict = {"type": "int", "bits": 32}
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, IntType)
    assert recap_type.type_ == "int"
    assert recap_type.bits == 32


def test_float_type():
    test_dict = {"type": "float", "bits": 64}
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, FloatType)
    assert recap_type.type_ == "float"
    assert recap_type.bits == 64


def test_string_type():
    test_dict = {"type": "string", "bytes": 32}
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, StringType)
    assert recap_type.type_ == "string"
    assert recap_type.bytes_ == 32


def test_bytes_type():
    test_dict = {"type": "bytes", "bytes": 32}
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, BytesType)
    assert recap_type.type_ == "bytes"
    assert recap_type.bytes_ == 32


def test_list_type():
    test_dict = {"type": "list", "values": {"type": "int", "bits": 32}}
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, ListType)
    assert recap_type.type_ == "list"
    assert isinstance(recap_type.values, IntType)
    assert recap_type.values.type_ == "int"
    assert recap_type.values.bits == 32


def test_map_type():
    test_dict = {
        "type": "map",
        "keys": {"type": "int", "bits": 32},
        "values": {"type": "string", "bytes": 32},
    }
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, MapType)
    assert recap_type.type_ == "map"
    assert isinstance(recap_type.keys, IntType)
    assert recap_type.keys.type_ == "int"
    assert recap_type.keys.bits == 32
    assert isinstance(recap_type.values, StringType)
    assert recap_type.values.type_ == "string"
    assert recap_type.values.bytes_ == 32


def test_struct_type():
    test_dict = {
        "type": "struct",
        "fields": [{"type": "int", "bits": 32}, {"type": "string", "bytes": 32}],
    }
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, StructType)
    assert recap_type.type_ == "struct"
    for field in recap_type.fields:
        if isinstance(field, IntType):
            assert field.type_ == "int"
            assert field.bits == 32
        elif isinstance(field, StringType):
            assert field.type_ == "string"
            assert field.bytes_ == 32


def test_enum_type():
    test_dict = {"type": "enum", "symbols": ["RED", "GREEN", "BLUE"]}
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, EnumType)
    assert recap_type.type_ == "enum"
    assert recap_type.symbols == ["RED", "GREEN", "BLUE"]


@pytest.mark.parametrize(
    "test_dict",
    [
        {
            "type": "union",
            "types": [{"type": "int", "bits": 32}, {"type": "string", "bytes": 32}],
        },
        {
            "type": "union",
            "types": ["int32", "string32"],
        },
        {
            "type": "union",
            "types": ["int32", {"type": "string", "bytes": 32}],
        },
    ],
)
def test_union_type(test_dict):
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, UnionType)
    assert recap_type.type_ == "union"
    for type_ in recap_type.types:
        if isinstance(type_, IntType):
            assert type_.type_ == "int"
            assert type_.bits == 32
        elif isinstance(type_, StringType):
            assert type_.type_ == "string"
            assert type_.bytes_ == 32


aliases = [
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "float16",
    "float32",
    "float64",
    "string32",
    "string64",
    "bytes32",
    "bytes64",
    "uuid",
    "decimal128",
    "decimal256",
    "duration64",
    "interval128",
    "time32",
    "time64",
    "timestamp64",
    "date32",
    "date64",
]


@pytest.mark.parametrize("alias", aliases)
def test_from_dict_aliases(alias):
    recap_type_registry = RecapTypeRegistry()
    recap_type = from_dict({"type": alias}, recap_type_registry)
    assert isinstance(recap_type, ProxyType)
    assert recap_type.resolve() == recap_type_registry.from_alias(alias)


def test_from_dict_raises_for_missing_type():
    with pytest.raises(AssertionError):
        from_dict({"alias": "alias"})


def test_from_dict_self_referencing_structure():
    # define the test_dict with the self-referencing structure
    test_dict = {
        "type": "struct",
        "fields": [
            {"type": "int", "bits": 32},
            {"type": "build.recap.SelfReferencingStructure"},
        ],
        "alias": "build.recap.SelfReferencingStructure",
    }

    # Create a self-referencing RecapType
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, StructType)
    assert recap_type.type_ == "struct"
    assert len(recap_type.fields) == 2
    if isinstance(recap_type.fields[0], IntType):
        assert recap_type.fields[0].type_ == "int"
        assert recap_type.fields[0].bits == 32
    elif isinstance(recap_type.fields[1], ProxyType):
        assert recap_type.fields[1].type_ == "proxy"
        assert recap_type.fields[1].alias == "build.recap.SelfReferencingStructure"
        # Resolve the ProxyType and check it equals the original RecapType
        assert recap_type.fields[1].resolve() == recap_type


def test_from_dict_alias_with_attribute_override():
    # Define a dictionary with a struct RecapType that includes a
    # self-reference with an attribute override
    test_dict = {
        "type": "struct",
        "fields": [
            {"type": "int", "bits": 32, "alias": "build.recap.MyInt"},
            {"type": "build.recap.MyInt", "signed": False},
        ],
    }

    # Create a self-referencing RecapType with attribute overrides
    recap_type = from_dict(test_dict)
    assert isinstance(recap_type, StructType)
    assert recap_type.type_ == "struct"
    for field in recap_type.fields:
        if isinstance(field, IntType):
            assert field.type_ == "int"
            assert field.bits == 32
            assert field.signed is True
        elif isinstance(field, ProxyType):
            assert field.type_ == "proxy"
            assert field.alias == "build.recap.MyInt"
            # Resolve the ProxyType and check it equals the original RecapType
            resolved_type = field.resolve()
            assert isinstance(resolved_type, IntType)
            assert resolved_type.type_ == "int"
            assert resolved_type.bits == 32
            assert resolved_type.signed is False


def test_from_dict_uuid_logical_type():
    logical_type_dict = {
        "type": "string",
        "logical": "build.recap.UUID",
        "bytes": 36,
        "variable": False,
    }
    recap_type = from_dict(logical_type_dict)
    assert isinstance(recap_type, StringType)
    assert recap_type.logical == logical_type_dict["logical"]
    assert recap_type.bytes_ == logical_type_dict["bytes"]
    assert recap_type.variable == logical_type_dict["variable"]


def test_from_dict_decimal128_logical_type():
    logical_type_dict = {
        "type": "bytes",
        "logical": "build.recap.Decimal",
        "bytes": 16,
        "variable": False,
        "precision": 28,
        "scale": 14,
    }
    recap_type = from_dict(logical_type_dict)
    assert isinstance(recap_type, BytesType)
    assert recap_type.logical == logical_type_dict["logical"]
    assert recap_type.bytes_ == logical_type_dict["bytes"]
    assert recap_type.variable == logical_type_dict["variable"]
    assert recap_type.extra_attrs["precision"] == logical_type_dict["precision"]
    assert recap_type.extra_attrs["scale"] == logical_type_dict["scale"]


def test_from_dict_duration64_logical_type():
    logical_type_dict = {
        "type": "int",
        "logical": "build.recap.Duration",
        "bits": 64,
        "signed": True,
        "unit": "millisecond",
    }
    recap_type = from_dict(logical_type_dict)
    assert isinstance(recap_type, IntType)
    assert recap_type.logical == logical_type_dict["logical"]
    assert recap_type.bits == logical_type_dict["bits"]
    assert recap_type.signed == logical_type_dict["signed"]
    assert recap_type.extra_attrs["unit"] == logical_type_dict["unit"]


def test_from_dict_time64_logical_type():
    logical_type_dict = {
        "type": "int",
        "logical": "build.recap.Time",
        "bits": 64,
        "signed": True,
        "unit": "second",
    }
    recap_type = from_dict(logical_type_dict)
    assert isinstance(recap_type, IntType)
    assert recap_type.logical == logical_type_dict["logical"]
    assert recap_type.bits == logical_type_dict["bits"]
    assert recap_type.signed == logical_type_dict["signed"]
    assert recap_type.extra_attrs["unit"] == logical_type_dict["unit"]


def test_from_dict_timestamp64_logical_type():
    logical_type_dict = {
        "type": "int",
        "logical": "build.recap.Timestamp",
        "bits": 64,
        "signed": True,
        "unit": "millisecond",
    }
    recap_type = from_dict(logical_type_dict)
    assert isinstance(recap_type, IntType)
    assert recap_type.logical == logical_type_dict["logical"]
    assert recap_type.bits == logical_type_dict["bits"]
    assert recap_type.signed == logical_type_dict["signed"]
    assert recap_type.extra_attrs["unit"] == logical_type_dict["unit"]


def test_from_dict_date64_logical_type():
    logical_type_dict = {
        "type": "int",
        "logical": "build.recap.Date",
        "bits": 64,
        "signed": True,
        "unit": "day",
    }
    recap_type = from_dict(logical_type_dict)
    assert isinstance(recap_type, IntType)
    assert recap_type.logical == logical_type_dict["logical"]
    assert recap_type.bits == logical_type_dict["bits"]
    assert recap_type.signed == logical_type_dict["signed"]
    assert recap_type.extra_attrs["unit"] == logical_type_dict["unit"]


def test_from_dict_interval_logical_type():
    logical_type_dict = {
        "type": "bytes",
        "logical": "build.recap.Interval",
        "bytes": 12,
        "variable": False,
    }
    recap_type = from_dict(logical_type_dict)
    assert isinstance(recap_type, BytesType)
    assert recap_type.logical == logical_type_dict["logical"]
    assert recap_type.bytes_ == logical_type_dict["bytes"]
    assert recap_type.variable == logical_type_dict["variable"]


def test_from_dict_complex_struct_type():
    input_dict = {
        "type": "struct",
        "fields": [
            {
                "type": "list",
                "values": {
                    "type": "map",
                    "keys": {"type": "int", "bits": 32},
                    "values": {"type": "string", "bytes": 50},
                },
            }
        ],
    }
    result = from_dict(input_dict)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], ListType)
    assert isinstance(result.fields[0].values, MapType)
    assert isinstance(result.fields[0].values.keys, IntType)
    assert result.fields[0].values.keys.bits == 32
    assert isinstance(result.fields[0].values.values, StringType)
    assert result.fields[0].values.values.bytes_ == 50


def test_from_dict_complex_union_type():
    input_dict = {
        "type": "union",
        "types": [
            {"type": "int", "bits": 32},
            {"type": "struct", "fields": [{"type": "bool"}]},
        ],
    }
    result = from_dict(input_dict)
    assert isinstance(result, UnionType)
    assert len(result.types) == 2
    assert isinstance(result.types[0], IntType)
    assert result.types[0].bits == 32
    assert isinstance(result.types[1], StructType)
    assert isinstance(result.types[1].fields[0], BoolType)


def test_from_dict_missing_type():
    with pytest.raises(AssertionError):
        input_dict = {}
        from_dict(input_dict)


def test_from_dict_missing_required_field_for_int():
    with pytest.raises(ValueError):
        input_dict = {"type": "int"}
        from_dict(input_dict)


def test_from_dict_from_dict_with_union():
    # Define a dictionary representing a union of 'null' and a complex type
    type_dict = {
        "type": [
            "null",
            {
                "type": "map",
                "keys": {"type": "int", "bits": 32},
                "values": {"type": "string", "bytes": 4},
            },
        ]
    }

    # Call from_dict and verify the resulting RecapType
    recap_type = from_dict(type_dict)

    assert isinstance(recap_type, UnionType)
    assert len(recap_type.types) == 2
    assert isinstance(recap_type.types[0], NullType)
    assert isinstance(recap_type.types[1], MapType)
    assert isinstance(recap_type.types[1].keys, IntType)
    assert recap_type.types[1].keys.bits == 32
    assert isinstance(recap_type.types[1].values, StringType)
    assert recap_type.types[1].values.bytes_ == 4


def test_from_dict_with_union_alias():
    # Define a dictionary representing a union of 'null' and 'timestamp64'
    type_dict = {
        "type": [
            "null",
            "timestamp64",
        ]
    }

    # Call from_dict and verify the resulting RecapType
    recap_type = from_dict(type_dict)

    assert isinstance(recap_type, UnionType)
    assert len(recap_type.types) == 2
    assert isinstance(recap_type.types[0], NullType)
    assert isinstance(recap_type.types[1], ProxyType)
    assert recap_type.types[1].alias == "timestamp64"
    resolved_type = recap_type.types[1].resolve()
    assert isinstance(resolved_type, IntType)
    assert resolved_type.logical == "build.recap.Timestamp"
    assert resolved_type.bits == 64
    assert resolved_type.signed is True
    assert resolved_type.extra_attrs["unit"] == "millisecond"


def test_from_dict_struct_with_string_field_no_bytes_set():
    input_dict = {
        "type": "struct",
        "fields": [
            {
                "type": "string",
            }
        ],
    }
    result = from_dict(input_dict)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], StringType)
    assert result.fields[0].bytes_ == 65_536


def test_from_dict_struct_with_bytes_field_no_bytes_set():
    input_dict = {
        "type": "struct",
        "fields": [
            {
                "type": "bytes",
            }
        ],
    }
    result = from_dict(input_dict)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], BytesType)
    assert result.fields[0].bytes_ == 65_536


def test_from_dict_optional_field():
    # Define a dictionary with a struct RecapType that includes an optional field
    test_dict = {
        "type": "struct",
        "fields": [
            {
                "name": "field1",
                "type": "int",
                "bits": 32,
            },
            {
                "name": "field2",
                "type": "float",
                "bits": 64,
                "optional": True,
            },
        ],
    }

    # Create a RecapType from the dictionary
    recap_type = from_dict(test_dict)

    # Verify the RecapType
    assert isinstance(recap_type, StructType)
    assert recap_type.type_ == "struct"
    assert len(recap_type.fields) == 2

    # Verify the first field
    field1 = recap_type.fields[0]
    assert field1.extra_attrs["name"] == "field1"
    assert isinstance(field1, IntType)
    assert field1.type_ == "int"
    assert field1.bits == 32

    # Verify the second field
    field2 = recap_type.fields[1]
    assert field2.extra_attrs.get("name") == "field2"
    assert field2.extra_attrs["default"] is None
    assert isinstance(field2, UnionType)
    assert isinstance(field2.types[0], NullType)
    assert isinstance(field2.types[1], FloatType)
    assert field2.types[1].type_ == "float"
    assert field2.types[1].bits == 64


def test_from_dict_alias_with_default_and_doc():
    # Verify that defaults in aliases are excluded
    test_dict = {
        "type": "struct",
        "fields": [
            {
                "name": "field1",
                "alias": "build.recap.Int32",
                "doc": "This is a test",
                "type": "int",
                "bits": 32,
                "default": -1,
            },
            {
                "name": "field2",
                "type": "build.recap.Int32",
            },
        ],
    }

    # Create a RecapType from the dictionary
    recap_type = from_dict(test_dict)

    # Verify the RecapType
    assert isinstance(recap_type, StructType)
    assert recap_type.type_ == "struct"
    assert len(recap_type.fields) == 2

    # Verify the first field
    field1 = recap_type.fields[0]
    assert field1.extra_attrs["name"] == "field1"
    assert isinstance(field1, IntType)
    assert field1.alias == "build.recap.Int32"
    assert field1.type_ == "int"
    assert field1.doc == "This is a test"
    assert field1.bits == 32
    assert field1.extra_attrs.get("default") == -1

    # Verify the second field
    field2 = recap_type.fields[1]
    assert field2.extra_attrs.get("name") == "field2"
    assert field2.extra_attrs.get("default") is None
    assert isinstance(field2, ProxyType)
    resolved_type = field2.resolve()
    assert isinstance(resolved_type, IntType)
    assert resolved_type.type_ == "int"
    assert resolved_type.bits == 32
    assert resolved_type.alias is None
    assert resolved_type.doc is None
    assert resolved_type.extra_attrs.get("name") == "field2"
    assert resolved_type.extra_attrs.get("default") is None


def test_from_dict_alias_with_no_default():
    # Verify that defaults in alias references are included
    test_dict = {
        "type": "struct",
        "fields": [
            {
                "name": "field1",
                "alias": "build.recap.Int32",
                "type": "int",
                "bits": 32,
            },
            {
                "name": "field2",
                "type": "build.recap.Int32",
                "default": -1,
            },
        ],
    }

    # Create a RecapType from the dictionary
    recap_type = from_dict(test_dict)

    # Verify the RecapType
    assert isinstance(recap_type, StructType)
    assert recap_type.type_ == "struct"
    assert len(recap_type.fields) == 2

    # Verify the first field
    field1 = recap_type.fields[0]
    assert field1.extra_attrs["name"] == "field1"
    assert isinstance(field1, IntType)
    assert field1.alias == "build.recap.Int32"
    assert field1.type_ == "int"
    assert field1.bits == 32
    assert field1.extra_attrs.get("default") is None

    # Verify the second field
    field2 = recap_type.fields[1]
    assert field2.extra_attrs.get("name") == "field2"
    assert field2.extra_attrs.get("default") == -1
    assert isinstance(field2, ProxyType)
    resolved_type = field2.resolve()
    assert isinstance(resolved_type, IntType)
    assert resolved_type.type_ == "int"
    assert resolved_type.bits == 32
    assert resolved_type.alias is None
    assert resolved_type.extra_attrs.get("name") == "field2"
    assert resolved_type.extra_attrs.get("default") == -1


@pytest.mark.parametrize(
    "input_value, expected_dict, clean, alias",
    [
        (
            NullType(),
            {
                "type": "null",
                "alias": None,
                "doc": None,
                "logical": None,
            },
            False,
            False,
        ),
        (
            NullType(),
            "null",
            True,
            False,
        ),
        (
            BoolType(),
            "bool",
            True,
            False,
        ),
        (
            IntType(bits=32),
            {
                "type": "int",
                "bits": 32,
                "signed": True,
                "alias": None,
                "doc": None,
                "logical": None,
            },
            False,
            False,
        ),
        (
            IntType(bits=32),
            {"type": "int", "bits": 32},
            True,
            False,
        ),
        (
            IntType(bits=32, signed=False),
            {"type": "int", "bits": 32, "signed": False},
            True,
            False,
        ),
        (
            FloatType(bits=32),
            {"type": "float", "bits": 32},
            True,
            False,
        ),
        (
            StringType(),
            "string",
            True,
            False,
        ),
        (
            StringType(bytes_=50),
            {"type": "string", "bytes": 50},
            True,
            False,
        ),
        (
            BytesType(),
            "bytes",
            True,
            False,
        ),
        (
            BytesType(bytes_=50),
            {"type": "bytes", "bytes": 50},
            True,
            False,
        ),
        (
            EnumType(symbols=["foo", "bar"]),
            {"type": "enum", "symbols": ["foo", "bar"]},
            True,
            False,
        ),
        (
            ListType(values=IntType(bits=32)),
            {"type": "list", "values": {"type": "int", "bits": 32}},
            True,
            False,
        ),
        (
            ListType(
                values=IntType(bits=32),
                length=10,
            ),
            {
                "type": "list",
                "values": {"type": "int", "bits": 32},
                "length": 10,
            },
            True,
            False,
        ),
        (
            MapType(
                keys=IntType(bits=32),
                values=StringType(bytes_=50),
            ),
            {
                "type": "map",
                "keys": {"type": "int", "bits": 32},
                "values": {"type": "string", "bytes": 50},
            },
            True,
            False,
        ),
        (
            UnionType(
                types=[
                    IntType(bits=32),
                    StringType(bytes_=50),
                ]
            ),
            [
                {"type": "int", "bits": 32},
                {"type": "string", "bytes": 50},
            ],
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    IntType(bits=32),
                    StringType(bytes_=50),
                ]
            ),
            {
                "type": "struct",
                "fields": [
                    {"type": "int", "bits": 32},
                    {"type": "string", "bytes": 50},
                ],
            },
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    UnionType(
                        default=None,
                        types=[
                            NullType(),
                            IntType(bits=32),
                            StringType(bytes_=50),
                        ],
                    )
                ]
            ),
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "union",
                        "types": [
                            {"type": "int", "bits": 32},
                            {"type": "string", "bytes": 50},
                        ],
                        "optional": True,
                    }
                ],
            },
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    ListType(
                        values=MapType(
                            keys=IntType(bits=32), values=StringType(bytes_=50)
                        )
                    )
                ]
            ),
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "list",
                        "values": {
                            "type": "map",
                            "keys": {"type": "int", "bits": 32},
                            "values": {"type": "string", "bytes": 50},
                        },
                    }
                ],
            },
            True,
            False,
        ),
        (
            ProxyType("test_proxy", RecapTypeRegistry()),
            "test_proxy",
            True,
            False,
        ),
        (
            IntType(bits=32),
            "int32",
            True,
            True,
        ),
        (
            BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=56,
                scale=28,
            ),
            "decimal256",
            True,
            True,
        ),
        (
            StructType(
                fields=[
                    StringType(
                        name="foo",
                        bytes_=10,
                        variable=False,
                        default="bar",
                    ),
                ],
            ),
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "string",
                        "bytes": 10,
                        "variable": False,
                        "default": "bar",
                        "name": "foo",
                    },
                ],
            },
            True,
            True,
        ),
        (
            StructType(
                fields=[
                    UnionType(
                        types=[
                            NullType(),
                            StringType(bytes_=50),
                            IntType(bits=32),
                        ],
                        default=None,
                    ),
                ],
            ),
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "union",
                        "types": [
                            {"type": "string", "bytes": 50},
                            {"type": "int", "bits": 32},
                        ],
                        "optional": True,
                    }
                ],
            },
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    UnionType(
                        types=[
                            NullType(),
                            StringType(bytes_=50),
                            IntType(bits=32),
                        ],
                        default="some_string",
                    ),
                ],
            ),
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "union",
                        "types": [
                            {"type": "string", "bytes": 50},
                            {"type": "int", "bits": 32},
                        ],
                        "optional": True,
                        "default": "some_string",
                    }
                ],
            },
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    UnionType(
                        types=[
                            NullType(),
                            StringType(bytes_=50),
                            IntType(bits=32),
                        ],
                        default=None,
                    ),
                ],
            ),
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "union",
                        "types": [
                            {"type": "string", "bytes": 50},
                            "int32",
                        ],
                        "optional": True,
                    }
                ],
            },
            True,
            True,
        ),
        (
            StructType(
                fields=[
                    UnionType(
                        types=[
                            NullType(),
                            IntType(bits=32),
                        ],
                    ),
                ],
            ),
            {
                "type": "struct",
                "fields": [
                    # Union syntactic sugar
                    [
                        "null",
                        "int32",
                    ],
                ],
            },
            True,
            True,
        ),
        (
            StructType(
                fields=[
                    UnionType(
                        types=[
                            NullType(),
                            StringType(bytes_=50),
                            IntType(bits=32),
                        ],
                        default=None,
                        alias="recap.build.TestUnion",
                    ),
                ],
            ),
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "union",
                        "alias": "recap.build.TestUnion",
                        "types": [
                            {"type": "string", "bytes": 50},
                            "int32",
                        ],
                        "optional": True,
                    }
                ],
            },
            True,
            True,
        ),
        (
            StructType(
                fields=[
                    StringType(
                        name="primary_phone",
                        alias="build.recap.Phone",
                        bytes_=10,
                    ),
                    UnionType(
                        name="secondary_phone",
                        types=[
                            NullType(),
                            ProxyType(
                                alias="build.recap.Phone",
                                registry=RecapTypeRegistry(),
                            ),
                        ],
                        default=None,
                    ),
                ],
            ),
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "string",
                        "bytes": 10,
                        "name": "primary_phone",
                        "alias": "build.recap.Phone",
                    },
                    {
                        "type": "build.recap.Phone",
                        "name": "secondary_phone",
                        "optional": True,
                    },
                ],
            },
            True,
            True,
        ),
    ],
)
def test_to_dict(
    input_value: RecapType,
    expected_dict: dict,
    clean: bool,
    alias: bool,
) -> None:
    assert to_dict(input_value, clean, alias) == expected_dict


@pytest.mark.parametrize(
    "input_value, clean, alias",
    [
        (
            NullType(),
            False,
            False,
        ),
        (
            NullType(),
            True,
            False,
        ),
        (
            BoolType(),
            True,
            False,
        ),
        (
            IntType(bits=32),
            False,
            False,
        ),
        (
            IntType(bits=32),
            True,
            False,
        ),
        (
            IntType(bits=32, signed=False),
            True,
            False,
        ),
        (
            FloatType(bits=32),
            True,
            False,
        ),
        (
            StringType(),
            True,
            False,
        ),
        (
            StringType(bytes_=50),
            True,
            False,
        ),
        (
            BytesType(),
            True,
            False,
        ),
        (
            BytesType(bytes_=50),
            True,
            False,
        ),
        (
            EnumType(symbols=["foo", "bar"]),
            True,
            False,
        ),
        (
            ListType(values=IntType(bits=32)),
            True,
            False,
        ),
        (
            ListType(
                values=IntType(bits=32),
                length=10,
            ),
            True,
            False,
        ),
        (
            MapType(
                keys=IntType(bits=32),
                values=StringType(bytes_=50),
            ),
            True,
            False,
        ),
        (
            UnionType(
                types=[
                    IntType(bits=32),
                    StringType(bytes_=50),
                ]
            ),
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    IntType(bits=32),
                    StringType(bytes_=50),
                ]
            ),
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    UnionType(
                        default=None,
                        types=[
                            NullType(),
                            IntType(bits=32),
                            StringType(bytes_=50),
                        ],
                    )
                ]
            ),
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    ListType(
                        values=MapType(
                            keys=IntType(bits=32), values=StringType(bytes_=50)
                        )
                    )
                ]
            ),
            True,
            False,
        ),
        (
            ProxyType("test_proxy", RecapTypeRegistry()),
            True,
            False,
        ),
        (
            ProxyType("test_proxy", RecapTypeRegistry()),
            True,
            True,
        ),
        (
            IntType(bits=32),
            True,
            False,
        ),
        (
            BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=56,
                scale=28,
            ),
            True,
            False,
        ),
        (
            StructType(
                fields=[
                    StringType(
                        name="foo",
                        bytes_=10,
                        variable=False,
                        default="bar",
                    ),
                ],
            ),
            True,
            False,
        ),
    ],
)
def test_to_from_dict(
    input_value: RecapType,
    clean: bool,
    alias: bool,
):
    assert from_dict(to_dict(input_value, clean, alias)) == input_value


@pytest.mark.parametrize(
    "input_value, clean, alias",
    [
        (
            {
                "type": "null",
                "alias": None,
                "doc": None,
                "logical": None,
            },
            False,
            False,
        ),
        ("null", True, False),
        ("bool", True, False),
        (
            {
                "type": "int",
                "bits": 32,
                "signed": True,
                "alias": None,
                "doc": None,
                "logical": None,
            },
            False,
            False,
        ),
        ({"type": "int", "bits": 32}, True, False),
        ({"type": "int", "bits": 32, "signed": False}, True, False),
        ({"type": "float", "bits": 32}, True, False),
        ("string", True, False),
        ({"type": "string", "bytes": 50}, True, False),
        ("bytes", True, False),
        ({"type": "bytes", "bytes": 50}, True, False),
        ({"type": "enum", "symbols": ["foo", "bar"]}, True, False),
        ({"type": "list", "values": {"type": "int", "bits": 32}}, True, False),
        (
            {
                "type": "list",
                "values": {"type": "int", "bits": 32},
                "length": 10,
            },
            True,
            False,
        ),
        (
            {
                "type": "map",
                "keys": {"type": "int", "bits": 32},
                "values": {"type": "string", "bytes": 50},
            },
            True,
            False,
        ),
        (
            [
                {"type": "int", "bits": 32},
                {"type": "string", "bytes": 50},
            ],
            True,
            False,
        ),
        (
            {
                "type": "struct",
                "fields": [
                    {"type": "int", "bits": 32},
                    {"type": "string", "bytes": 50},
                ],
            },
            True,
            False,
        ),
        (
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "union",
                        "types": [
                            {"type": "int", "bits": 32},
                            {"type": "string", "bytes": 50},
                        ],
                        "optional": True,
                    }
                ],
            },
            True,
            False,
        ),
        (
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "list",
                        "values": {
                            "type": "map",
                            "keys": {"type": "int", "bits": 32},
                            "values": {"type": "string", "bytes": 50},
                        },
                    }
                ],
            },
            True,
            False,
        ),
        ("test_proxy", True, True),
        ("test_proxy", True, False),
        ("int32", True, True),
        ("int32", True, False),
        ("decimal256", True, True),
        ("decimal256", True, False),
        (
            {
                "type": "struct",
                "fields": [
                    {
                        "type": "string",
                        "bytes": 10,
                        "variable": False,
                        "default": "bar",
                        "name": "foo",
                    },
                ],
            },
            True,
            True,
        ),
    ],
)
def test_from_to_dict(
    input_value: dict,
    clean: bool,
    alias: bool,
) -> None:
    assert to_dict(from_dict(input_value), clean, alias) == input_value


def test_to_dict_compact_union():
    union_type = UnionType(
        types=[
            IntType(bits=30),
            StringType(bytes_=50),
        ]
    )
    assert to_dict(union_type) == [
        {"type": "int", "bits": 30},
        {"type": "string", "bytes": 50},
    ]


def test_to_dict_compact_type():
    assert to_dict(IntType(bits=32)) == "int32"


@pytest.mark.parametrize(
    "test_input, expected",
    [
        # Valid cases
        ("foo.bar.Baz", True),
        ("foo.Baz", True),
        ("foo123_.blah", True),
        ("_foo.blah", True),
        ("bar123___.___b123lah___43", True),
        ("foo.bar.baz.blah", True),
        # Invalid cases
        ("foo", False),
        ("123foo", False),
        ("123.blah", False),
        ("123blah.blah", False),
        ("blah.123blah", False),
        (".blah", False),
        ("blah.", False),
        ("foo..blah", False),
        ("foo.blah.", False),
    ],
)
def test_alias_regex(test_input, expected):
    assert (re.fullmatch(ALIAS_REGEX, test_input) is not None) == expected


@pytest.mark.parametrize(
    "recap_type",
    [
        # Valid cases
        pytest.param(BoolType(alias="foo.bar.Baz"), id="valid:foo.bar.Baz"),
        pytest.param(BoolType(alias="foo.Baz"), id="valid:foo.Baz"),
        pytest.param(BoolType(alias="foo123_.blah"), id="valid:foo123_.blah"),
        pytest.param(BoolType(alias="foo_.blah123"), id="valid:foo_.blah123"),
        pytest.param(BoolType(alias="_foo.blah"), id="valid:_foo.blah"),
        pytest.param(
            BoolType(alias="bar123___.___b123lah___43"),
            id="valid:bar123___.___b123lah___43",
        ),
        # Invalid cases
        pytest.param(
            BoolType(alias="foo"),
            id="invalid:foo",
            marks=pytest.mark.xfail,
        ),
        pytest.param(
            BoolType(alias="123foo"),
            id="invalid:123foo",
            marks=pytest.mark.xfail,
        ),
        pytest.param(
            BoolType(alias=".blah"),
            id="invalid:.blah",
            marks=pytest.mark.xfail,
        ),
        pytest.param(
            BoolType(alias="blah."),
            id="invalid:blah.",
            marks=pytest.mark.xfail,
        ),
        pytest.param(
            BoolType(alias="foo..blah"),
            id="invalid:foo..blah",
            marks=pytest.mark.xfail,
        ),
        pytest.param(
            BoolType(alias="foo.blah."),
            id="invalid:foo..blah",
            marks=pytest.mark.xfail,
        ),
        pytest.param(
            BoolType(alias=""),
            id="invalid:foo..blah",
            marks=pytest.mark.xfail,
        ),
        pytest.param(
            BoolType(alias=" foo.Baz"),
            id="valid:foo.Baz",
            marks=pytest.mark.xfail,
        ),
        pytest.param(
            BoolType(alias="foo.Baz "),
            id="valid:foo.Baz",
            marks=pytest.mark.xfail,
        ),
    ],
)
def test_register_alias(recap_type):
    registry = RecapTypeRegistry()
    registry.register_alias(recap_type)


def test_int_type_validate():
    invalid_int1 = IntType(-1)
    invalid_int2 = IntType(2_147_483_648)

    with pytest.raises(ValueError):
        invalid_int1.validate()
    with pytest.raises(ValueError):
        invalid_int2.validate()


def test_float_type_validate():
    invalid_float1 = FloatType(-1)
    invalid_float2 = FloatType(2_147_483_648)

    with pytest.raises(ValueError):
        invalid_float1.validate()
    with pytest.raises(ValueError):
        invalid_float2.validate()


def test_string_type_validate():
    invalid_string1 = StringType(-1)
    invalid_string2 = StringType(9_223_372_036_854_775_808)

    with pytest.raises(ValueError):
        invalid_string1.validate()
    with pytest.raises(ValueError):
        invalid_string2.validate()


def test_bytes_type_validate():
    invalid_bytes1 = BytesType(-1)
    invalid_bytes2 = BytesType(9_223_372_036_854_775_808)

    with pytest.raises(ValueError):
        invalid_bytes1.validate()
    with pytest.raises(ValueError):
        invalid_bytes2.validate()


def test_list_type_validate():
    invalid_list1 = ListType(StringType(65_536), length=-1)
    invalid_list2 = ListType(StringType(65_536), variable=False)
    invalid_list3 = ListType(StringType(), length=9_223_372_036_854_775_808)

    with pytest.raises(ValueError):
        invalid_list1.validate()
    with pytest.raises(ValueError):
        invalid_list2.validate()
    with pytest.raises(ValueError):
        invalid_list3.validate()


def test_map_type_validate():
    invalid_map_keys = MapType(StringType(9_223_372_036_854_775_808), IntType(32))
    invalid_map_values = MapType(StringType(65_536), IntType(2_147_483_648))

    with pytest.raises(ValueError):
        invalid_map_keys.validate()
    with pytest.raises(ValueError):
        invalid_map_values.validate()


def test_struct_type_validate():
    invalid_struct = StructType([StringType(65_536), IntType(2_147_483_648)])

    with pytest.raises(ValueError):
        invalid_struct.validate()



from __future__ import annotations

import copy
import inspect
from re import fullmatch
from typing import Any

ALIAS_REGEX = r"^[a-zA-Z_][a-zA-Z0-9_]*(?:\.[a-zA-Z_][a-zA-Z0-9_]*)+$"


class RecapType:
    """Base class for all Recap types."""

    def __init__(
        self,
        type_: str,
        logical: str | None = None,
        alias: str | None = None,
        doc: str | None = None,
        **extra_attrs,
    ):
        self.type_ = type_
        self.logical = logical
        self.alias = alias
        self.doc = doc
        self.extra_attrs = extra_attrs

    def make_nullable(self) -> UnionType:
        """
        Make a type nullable by wrapping it in a union with null. Promote the
        doc and default values to the union.

        The nested type is a copy of `self` with `doc` and `default` removed.

        :return: A union type with null and the original type. Doc and default
            are promoted to the union if they're set.
        """

        RecapTypeClass = self.__class__
        attrs = vars(self)
        # Filter private attributes
        attrs = {k: v for k, v in attrs.items() if not k.startswith("_")}
        attrs.pop("type_", None)
        # Move doc to the union type
        doc = attrs.pop("doc", None)
        # Unnest extra_attrs for copy or we end up with extra_attrs["extra_attrs"]
        extra_attrs = attrs.pop("extra_attrs", {})
        union_attrs = {
            "default": extra_attrs.pop("default", None),
            "doc": doc,
        }
        # Move field name to the union type
        if "name" in extra_attrs:
            union_attrs["name"] = extra_attrs.pop("name")
        type_copy = RecapTypeClass(**attrs, **extra_attrs)
        return UnionType([NullType(), type_copy], **union_attrs)

    def validate(self) -> None:
        # Default to valid type
        pass

    def __eq__(self, other):
        if type(self) is type(other):
            return (
                self.type_,
                self.logical,
                self.alias,
                self.doc,
                self.extra_attrs,
            ) == (other.type_, other.logical, other.alias, other.doc, other.extra_attrs)
        return False

    def __repr__(self):
        attrs = vars(self)
        attrs_str = ", ".join(f"{k}={v}" for k, v in attrs.items())
        return f"{self.__class__.__name__}({attrs_str})"


class NullType(RecapType):
    """Represents a null Recap type."""

    def __init__(self, **extra_attrs):
        super().__init__("null", **extra_attrs)


class BoolType(RecapType):
    """Represents a boolean Recap type."""

    def __init__(self, **extra_attrs):
        super().__init__("bool", **extra_attrs)


class IntType(RecapType):
    """Represents an integer Recap type."""

    def __init__(self, bits: int, signed: bool = True, **extra_attrs):
        super().__init__("int", **extra_attrs)
        self.bits = bits
        self.signed = signed

    def __eq__(self, other):
        return super().__eq__(other) and (self.bits, self.signed) == (
            other.bits,
            other.signed,
        )

    def validate(self) -> None:
        if self.bits < 1 or self.bits > 2_147_483_647:
            raise ValueError("bits must be between 1 and 2,147,483,647")


class FloatType(RecapType):
    """Represents a floating point Recap type."""

    def __init__(self, bits: int, **extra_attrs):
        super().__init__("float", **extra_attrs)
        self.bits = bits

    def __eq__(self, other):
        return super().__eq__(other) and self.bits == other.bits

    def validate(self) -> None:
        if self.bits < 1 or self.bits > 2_147_483_647:
            raise ValueError("bits must be between 1 and 2,147,483,647")


class StringType(RecapType):
    """Represents a string Recap type."""

    def __init__(self, bytes_: int = 65_536, variable: bool = True, **extra_attrs):
        super().__init__("string", **extra_attrs)
        self.bytes_ = bytes_
        self.variable = variable

    def __eq__(self, other):
        return super().__eq__(other) and (self.bytes_, self.variable) == (
            other.bytes_,
            other.variable,
        )

    def validate(self) -> None:
        if self.bytes_ < 1 or self.bytes_ > 9_223_372_036_854_775_807:
            raise ValueError("bytes must be between 1 and 9,223,372,036,854,775,807")


class BytesType(RecapType):
    """Represents a bytes Recap type."""

    def __init__(self, bytes_: int = 65_536, variable: bool = True, **extra_attrs):
        super().__init__("bytes", **extra_attrs)
        self.bytes_ = bytes_
        self.variable = variable

    def __eq__(self, other):
        return super().__eq__(other) and (self.bytes_, self.variable) == (
            other.bytes_,
            other.variable,
        )

    def validate(self) -> None:
        if self.bytes_ < 1 or self.bytes_ > 9_223_372_036_854_775_807:
            raise ValueError("bytes must be between 1 and 9,223,372,036,854,775,807")


class ListType(RecapType):
    """Represents a list Recap type."""

    def __init__(
        self,
        values: RecapType,
        length: int | None = None,
        variable: bool = True,
        **extra_attrs,
    ):
        super().__init__("list", **extra_attrs)
        self.values = values
        self.length = length
        self.variable = variable

    def __eq__(self, other):
        return super().__eq__(other) and (self.values, self.length, self.variable) == (
            other.values,
            other.length,
            other.variable,
        )

    def validate(self) -> None:
        if not self.variable and self.length is None:
            raise ValueError("Fixed length lists must have a length set")
        if self.length and (self.length < 0 or self.length > 9_223_372_036_854_775_807):
            raise ValueError(
                "List length must be between 0 and 9,223,372,036,854,775,807"
            )
        self.values.validate()


class MapType(RecapType):
    """Represents a map Recap type."""

    def __init__(self, keys: RecapType, values: RecapType, **extra_attrs):
        super().__init__("map", **extra_attrs)
        self.keys = keys
        self.values = values

    def __eq__(self, other):
        return super().__eq__(other) and (self.keys, self.values) == (
            other.keys,
            other.values,
        )

    def validate(self) -> None:
        self.keys.validate()
        self.values.validate()


class StructType(RecapType):
    """Represents a struct Recap type."""

    def __init__(self, fields: list[RecapType] | None = None, **extra_attrs):
        super().__init__("struct", **extra_attrs)
        self.fields = fields if fields is not None else []

    def __eq__(self, other):
        return super().__eq__(other) and self.fields == other.fields

    def validate(self) -> None:
        for field in self.fields:
            field.validate()


class EnumType(RecapType):
    """Represents an enum Recap type."""

    def __init__(self, symbols: list[str], **extra_attrs):
        super().__init__("enum", **extra_attrs)
        self.symbols = symbols

    def __eq__(self, other):
        return super().__eq__(other) and self.symbols == other.symbols


class UnionType(RecapType):
    """Represents a union Recap type."""

    def __init__(self, types: list[RecapType], **extra_attrs):
        super().__init__("union", **extra_attrs)
        self.types = types

    def __eq__(self, other):
        return super().__eq__(other) and self.types == other.types

    @property
    def is_optional(self) -> bool:
        """
        Returns True if the union is an optional type, that has...

        - a null type
        - a default value
        - a single non-null type

        :return: True if the union is an optional type
        """
        has_null = NullType() in self.types
        has_default = "default" in self.extra_attrs
        has_non_null = len(list(filter(lambda t: t != NullType(), self.types))) == 1
        return has_null and has_default and has_non_null


class ProxyType(RecapType):
    """Represents a proxy to an aliased Recap type."""

    def __init__(self, alias: str, registry: RecapTypeRegistry, **extra_attrs):
        super().__init__("proxy", **extra_attrs)
        self.alias = alias
        self.registry = registry
        self._resolved = None

    def resolve(self) -> RecapType:
        if self._resolved is None:
            self._resolved = copy.deepcopy(self.registry.from_alias(self.alias))
            # Apply attribute overrides
            for attr, value in self.extra_attrs.items():
                if hasattr(self._resolved, attr):
                    setattr(self._resolved, attr, value)
                else:
                    self._resolved.extra_attrs[attr] = value
        return self._resolved

    def __eq__(self, other):
        return super().__eq__(other) and self.alias == other.alias


class RecapTypeRegistry:
    """Class that handles Recap type registration and resolution."""

    def __init__(self):
        # Define and register built-in aliases
        self._type_registry: dict[str, RecapType] = copy.deepcopy(BUILTIN_ALIASES)

    def register_alias(self, recap_type: RecapType):
        alias = recap_type.alias
        if alias is None:
            raise ValueError("RecapType must have an alias.")
        if alias in self._type_registry:
            raise ValueError(f"Alias {recap_type.alias} is already used.")
        if fullmatch(ALIAS_REGEX, alias) is None:
            raise ValueError(f"Alias {alias} must be a fully qualified name.")
        recap_type = copy.deepcopy(recap_type)
        # Registry contains types without aliases so aliased types don't get
        # accidentally redefined when referenced.
        recap_type.alias = None
        # Do not include doc in aliases
        recap_type.doc = None
        # Do not include default in aliases
        recap_type.extra_attrs.pop("default", None)
        self._type_registry[alias] = recap_type

    def from_alias(self, alias: str) -> RecapType:
        try:
            return self._type_registry[alias]
        except KeyError:
            raise TypeError(f"No RecapType with alias {alias} found.")


def from_dict(
    type_dict: dict | list | str,
    registry: RecapTypeRegistry | None = None,
) -> RecapType:
    """
    Create a RecapType from a dictionary, list, or string. Lists are treated as
    unions and strings are treated as aliases or simple types (e.g. "null",
    "bool", "string", or "bytes").

    :param type_dict: A dictionary, list, or string representing a Recap type.
    :param registry: A RecapTypeRegistry to use for resolving aliases.
    :return: A RecapType.
    """

    if isinstance(type_dict, list):
        type_dict = {"type": "union", "types": type_dict}
    elif isinstance(type_dict, str):
        type_dict = {"type": type_dict}

    # Create a copy to avoid modifying the input dictionary
    type_dict = type_dict.copy()

    assert (
        type_dict.get("type") is not None
    ), "Type dictionary must have a 'type' field."

    # Support optional=True syntactic sugar
    if type_dict.pop("optional", False):
        # Special handling for reserved words in struct fields
        name = type_dict.pop("name", None)
        default = type_dict.pop("default", None)
        if type_dict["type"] != "union":
            type_dict = {
                "type": "union",
                "types": ["null", type_dict],
            }
        if "null" not in type_dict["types"]:
            type_dict["types"] = ["null"] + type_dict["types"]
        if name is not None:
            type_dict["name"] = name
        type_dict["default"] = default

    registry = registry or RecapTypeRegistry()
    type_name = type_dict.pop("type")

    if isinstance(type_name, list):
        # If type is a list, handle as a union
        union_types = []
        for t in type_name:
            if isinstance(t, dict):
                union_types.append(from_dict(t, registry))
            elif isinstance(t, str):
                union_types.append(from_dict({"type": t}, registry))
        recap_type = UnionType(union_types, **type_dict)
    elif isinstance(type_name, str):
        match type_name:
            case "null":
                recap_type = NullType(**type_dict)
            case "bool":
                recap_type = BoolType(**type_dict)
            case "int":
                if "bits" not in type_dict:
                    raise ValueError("'bits' attribute is required for 'int' type.")
                recap_type = IntType(**type_dict)
            case "float":
                if "bits" not in type_dict:
                    raise ValueError("'bits' attribute is required for 'float' type.")
                recap_type = FloatType(**type_dict)
            case "string":
                if "bytes" in type_dict:
                    type_dict["bytes_"] = type_dict.pop("bytes")
                recap_type = StringType(**type_dict)
            case "bytes":
                if "bytes" in type_dict:
                    type_dict["bytes_"] = type_dict.pop("bytes")
                recap_type = BytesType(**type_dict)
            case "list":
                if "values" not in type_dict:
                    raise ValueError("'values' attribute is required for 'list' type.")
                recap_type = ListType(
                    from_dict(type_dict.pop("values"), registry),
                    **type_dict,
                )
            case "map":
                if "keys" not in type_dict or "values" not in type_dict:
                    raise ValueError(
                        "'keys' and 'values' attributes are required for 'map' type."
                    )
                recap_type = MapType(
                    from_dict(type_dict.pop("keys"), registry),
                    from_dict(type_dict.pop("values"), registry),
                    **type_dict,
                )
            case "struct":
                type_dict["fields"] = [
                    from_dict(f, registry) for f in type_dict.get("fields", [])
                ]
                recap_type = StructType(**type_dict)
            case "enum":
                if "symbols" not in type_dict:
                    raise ValueError("'symbols' attribute is required for 'enum' type.")
                recap_type = EnumType(**type_dict)
            case "union":
                if "types" not in type_dict:
                    raise ValueError("'types' attribute is required for 'union' type.")
                recap_type = UnionType(
                    [from_dict(t, registry) for t in type_dict.pop("types")],
                    **type_dict,
                )
            case _:
                recap_type = ProxyType(type_name, registry, **type_dict)
    else:
        raise ValueError("'type' must be a string or list.")

    recap_type.validate()

    # If alias exists, register the created RecapType
    if recap_type.alias is not None and not isinstance(recap_type, ProxyType):
        registry.register_alias(recap_type)

    return recap_type


def to_dict(recap_type: RecapType, clean=True, alias=True) -> dict | list | str:
    """
    Convert a RecapType to a dictionary, list, or string.

    :param recap_type: A RecapType.
    :param clean: If True, remove defaults, and make the dictionary as compact
        as possible.
    :param alias: If True, replace types that match an alias with the alias
        name.
    :return: A dictionary, list, or string representing a Recap type.  Lists
        are returned if clean=True and recap_type is a union with no extra
        attributes. Strings are returned if clean=True and recap_type is a
        simple type (e.g. "null", "bool", "string", or "bytes") or alias (e.g.
        "int32", "uint64", "decimal256", etc).
    """

    type_dict = {
        "type": recap_type.type_,
        "alias": recap_type.alias,
        "logical": recap_type.logical,
        "doc": recap_type.doc,
        **recap_type.extra_attrs,
    }

    match recap_type:
        case IntType(bits=bits, signed=sign):
            type_dict.update({"bits": bits, "signed": sign})
        case FloatType(bits=bits):
            type_dict["bits"] = bits
        case StringType(bytes_=bytes_, variable=variable):
            type_dict.update({"bytes": bytes_, "variable": variable})
        case BytesType(bytes_=bytes_, variable=variable):
            type_dict.update({"bytes": bytes_, "variable": variable})
        case ListType(values=values, length=length, variable=variable):
            type_dict.update(
                {
                    "values": to_dict(values, clean, alias),
                    "length": length,
                    "variable": variable,
                }
            )
        case MapType(keys=keys, values=values):
            type_dict.update(
                {
                    "keys": to_dict(keys, clean, alias),
                    "values": to_dict(values, clean, alias),
                }
            )
        case StructType(fields=fields):
            type_dict["fields"] = [
                to_dict(
                    field,
                    clean,
                    alias,
                )
                for field in fields
            ]
        case EnumType(symbols=symbols):
            type_dict["symbols"] = symbols
        case UnionType(types=types):
            type_dict["types"] = [
                to_dict(
                    type_,
                    clean,
                    alias,
                )
                if isinstance(type_, RecapType)
                else type_
                for type_ in types
            ]
        case ProxyType():
            # Switch `alias` to `type` for ProxyTypes. This is because
            # ProxyType abuses teh "type" attribute to always be "proxy". It
            # uses "alias" as the actual type.
            type_dict["type"] = type_dict.pop("alias")
        case NullType() | BoolType():
            # These types just have a "type" attribute
            pass
        case _:
            raise ValueError(f"Unsupported type: {recap_type.type_}")

    # Replace concrete type definitions with aliases when possible
    type_dict = alias_dict(type_dict) if alias else type_dict

    # Remove defaults (including non-None defaults)
    type_dict = clean_dict(type_dict) if clean else type_dict

    return type_dict


def clean_dict(type_dict: dict | list | str) -> dict | list | str:
    """
    Remove defaults from a type dictionary, replace unions with lists when
    possible, and replace simple types with strings.

    ```
    {
        "type": "union",
        "types": [
            {
                "type": "int",
                "alias": None,
                "doc": None,
                "logical": None,
                "bits": 32,
                "signed": true,
            },
        ],
    }
    ```

    Would become `[{ "type": "int", "bits": 32}]`

    :param type_dict: A type dictionary, list, or string.
    :return: A cleaner, more compact type dictionary, list, or string.
    """

    if isinstance(type_dict, list):
        type_dict = {
            "type": "union",
            "types": type_dict,
        }
    elif isinstance(type_dict, dict) and "type" not in type_dict:
        raise ValueError(
            "'type' is a required field and was not found in the dictionary."
        )
    elif isinstance(type_dict, str):
        type_dict = {
            "type": type_dict,
        }

    type_name = type_dict.get("type")

    if isinstance(type_name, str):
        recap_type_class = TYPE_CLASSES.get(type_name, ProxyType)
        param_defaults = {}

        # Get defaults from all parent classes
        for cls in recap_type_class.__mro__[:-1]:  # Exclude 'object' class
            sig = inspect.signature(cls.__init__)

            param_defaults.update(
                {
                    # rstrip is for removing '_' from 'bytes_'
                    k.rstrip("_"): v.default
                    for k, v in sig.parameters.items()
                    if v.default is not inspect.Parameter.empty
                }
            )

        # Remove defaults from dictionary
        type_dict = {
            k: v
            for k, v in type_dict.items()
            if k not in param_defaults or param_defaults[k] != v
        }

        if type_name in ("list", "map"):
            assert "values" in type_dict, f"{type_name} must have a 'values' attribute"
            type_dict["values"] = clean_dict(type_dict["values"])

        if type_name == "map":
            type_dict["keys"] = clean_dict(type_dict["keys"])

        if "fields" in type_dict and type_name == "struct":
            type_dict["fields"] = [clean_dict(field) for field in type_dict["fields"]]

        if type_name == "union":
            type_dict["types"] = [clean_dict(t) for t in type_dict["types"]]
    elif isinstance(type_name, list):
        type_dict = {
            "type": "union",
            "types": [clean_dict(t) for t in type_dict["types"]],
        }

    # Support "optional" sytnactic sugar.
    # Shorten {"type": "union", "types": ["null", ...], "default": ...}
    # to {"type": "union", "types": [...], "optional": True}
    if (
        type_dict["type"] == "union"
        and "null" in type_dict["types"]
        and "default" in type_dict
    ):
        # Pop off standard attributes that should be kept whether the type is kept
        # as a union or converted to a single nested type.
        alias = type_dict.pop("alias", None)
        logical = type_dict.pop("logical", None)
        default = type_dict.pop("default")
        doc = type_dict.pop("doc", None)
        name = type_dict.pop("name", None)

        # Only do syntactic sugar if the union is a union with no other
        # (non-standard) extra attributes.
        if len(type_dict) == 2:
            type_dict["types"].remove("null")

            # If the type was ["null", <some type>], use <some type> instead
            # of a union of a single type. Don't do this if union is an alias or
            # logical type, since we promoting the nested type would change what
            # the alias/logical type means.
            if len(type_dict["types"]) == 1 and alias is None and logical is None:
                nested_type = type_dict["types"][0]
                if isinstance(nested_type, str):
                    nested_type = {"type": nested_type}
                type_dict = nested_type
            assert isinstance(type_dict, dict), "Expected type_dict to be a dict."
            attributes = {
                "alias": alias,
                "default": default,
                "doc": doc,
                "logical": logical,
                "name": name,
                "optional": True,
            }
            # Filter None values
            attributes = {k: v for k, v in attributes.items() if v is not None}
            # Add attributes to the type_dict.
            type_dict |= attributes

    # Shorten simple types {"type": "type"} to "type"
    if len(type_dict) == 1:
        return type_dict["type"]

    # Shorten {"type": "union", "types": ["type"]} to ["type"]
    if type_dict["type"] == "union" and len(type_dict) == 2:
        return type_dict["types"]

    return type_dict


def alias_dict(
    type_dict: dict[str, Any],
    registry: RecapTypeRegistry | None = None,
) -> dict[str, Any]:
    """
    Replaces concrete type definitions with aliases when possible.

    ```
    {
        "type": "int",
        "bits": 32,
        "signed": true,
    }
    ```

    Would become `{"type": "int32"}`.

    :param type_dict: A type dictionary.
    :param registry: A RecapTypeRegistry containing aliases.
    :return: A type dictionary with aliases.
    """

    # Types with a defined alias must always remain as concrete type definitions.
    if type_dict.get("alias") is not None:
        return type_dict

    registry = registry or RecapTypeRegistry()
    recap_type = from_dict(type_dict, registry)

    # If there's a matching alias, replace the type_dict with it
    for alias, alias_type in registry._type_registry.items():
        if recap_type == alias_type:
            return {"type": alias}

    # Otherwise, recurse on nested type_dicts
    for key in ["values", "keys", "fields", "types"]:
        if key in type_dict:
            if isinstance(type_dict[key], list):
                type_dict[key] = [
                    alias_dict(t, registry) if isinstance(t, dict) else t
                    for t in type_dict[key]
                ]
            elif isinstance(type_dict[key], dict):
                type_dict[key] = alias_dict(type_dict[key], registry)

    return type_dict


TYPE_CLASSES = {
    "bool": BoolType,
    "null": NullType,
    "int": IntType,
    "float": FloatType,
    "string": StringType,
    "bytes": BytesType,
    "list": ListType,
    "map": MapType,
    "struct": StructType,
    "enum": EnumType,
    "union": UnionType,
}


BUILTIN_ALIASES = {
    "int8": IntType(8, signed=True),
    "uint8": IntType(8, signed=False),
    "int16": IntType(16, signed=True),
    "uint16": IntType(16, signed=False),
    "int32": IntType(32, signed=True),
    "uint32": IntType(32, signed=False),
    "int64": IntType(64, signed=True),
    "uint64": IntType(64, signed=False),
    "float16": FloatType(16),
    "float32": FloatType(32),
    "float64": FloatType(64),
    "string32": StringType(bytes_=2_147_483_648, variable=True),
    "string64": StringType(bytes_=9_223_372_036_854_775_807, variable=True),
    "bytes32": BytesType(bytes_=2_147_483_648, variable=True),
    "bytes64": BytesType(bytes_=9_223_372_036_854_775_807, variable=True),
    "uuid": StringType(logical="build.recap.UUID", bytes_=36, variable=False),
    "decimal128": BytesType(
        logical="build.recap.Decimal",
        bytes_=16,
        variable=False,
        precision=28,
        scale=14,
    ),
    "decimal256": BytesType(
        logical="build.recap.Decimal",
        bytes_=32,
        variable=False,
        precision=56,
        scale=28,
    ),
    "duration64": IntType(
        logical="build.recap.Duration",
        bits=64,
        unit="millisecond",
    ),
    "interval128": BytesType(
        logical="build.recap.Interval",
        bytes_=16,
        variable=False,
        unit="millisecond",
    ),
    "time32": IntType(
        logical="build.recap.Time",
        bits=32,
        unit="second",
    ),
    "time64": IntType(
        logical="build.recap.Time",
        bits=64,
        unit="second",
    ),
    "timestamp64": IntType(
        logical="build.recap.Timestamp",
        bits=64,
        unit="millisecond",
        timezone="UTC",
    ),
    "date32": IntType(
        logical="build.recap.Date",
        bits=32,
        unit="day",
    ),
    "date64": IntType(
        logical="build.recap.Date",
        bits=64,
        unit="day",
    ),
}



import json
from typing import Any

from recap.types import (
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    RecapType,
    RecapTypeRegistry,
    StringType,
    StructType,
    UnionType,
)


class AvroConverter:
    def __init__(self) -> None:
        self.registry = RecapTypeRegistry()

    def from_recap(self, recap_type: RecapType) -> dict[str, Any]:
        """
        Convert a Recap type to an Avro schema.

        :param recap_type: The Recap type to convert.
        :return: The Avro schema.
        """

        avro_schema = self._from_recap(recap_type)
        if isinstance(avro_schema, str):
            avro_schema = {"type": avro_schema}
        return avro_schema

    def _from_recap(
        self,
        recap_type: RecapType,
    ) -> dict[str, Any] | str:
        """
        Convert a Recap type to an Avro schema. Can return a string if the
        schema is simple.

        :param recap_type: The Recap type to convert.
        :return: The Avro schema or simple string type.
        """

        avro_schema = {}

        if recap_type.doc:
            avro_schema["doc"] = recap_type.doc

        if name := recap_type.extra_attrs.get("name"):
            avro_schema["name"] = name

        if "default" in recap_type.extra_attrs:
            avro_schema["default"] = recap_type.extra_attrs["default"]

        if (alias := recap_type.alias) and "." in alias:
            # If the alias is not naked (built-in), define it in Avro. Built-in
            # types are always resolved (so "int32" always becomes "int",
            # "decimal256" always becomes a bytes with logicalType "decimal",
            # etc.)
            avro_schema["aliases"] = [alias]

        match recap_type:
            case RecapType(logical=str()):
                avro_schema |= self._from_recap_logical(recap_type)
            case NullType():
                avro_schema["type"] = "null"
            case BoolType():
                avro_schema["type"] = "boolean"
            case IntType(bits=int(bits), signed=True) if bits <= 32:
                avro_schema["type"] = "int"
            case IntType(bits=int(bits), signed=bool(signed)) if (
                bits <= 32 and not signed
            ) or (bits <= 64 and signed):
                avro_schema["type"] = "long"
            case IntType(bits=int(bits), signed=bool(signed)):
                precision = len(str(2 ** (bits - (1 if signed else 0)))) - 1
                avro_schema |= {
                    "type": "bytes",
                    "logicalType": "decimal",
                    "precision": precision,
                }
            case FloatType(bits=int(bits)) if bits <= 32:
                avro_schema["type"] = "float"
            case FloatType(bits=int(bits)) if bits <= 64:
                avro_schema["type"] = "double"
            case StringType(bytes_=int(bytes_)) if bytes_ <= 9_223_372_036_854_775_807:
                avro_schema["type"] = "string"
            case BytesType(
                bytes_=int(bytes_),
                variable=True,
            ) if bytes_ <= 9_223_372_036_854_775_807:
                avro_schema["type"] = "bytes"
            case BytesType(
                bytes_=int(bytes_),
                variable=False,
            ) if bytes_ <= 9_223_372_036_854_775_807:
                avro_schema["type"] = "fixed"
                avro_schema["size"] = bytes_
            case StructType(fields=list(struct_fields)):
                avro_schema["type"] = "record"
                record_fields = []
                for struct_field in struct_fields:
                    record_field = {}
                    field_type = self._from_recap(struct_field)
                    if isinstance(field_type, dict):
                        record_field |= field_type
                    elif isinstance(field_type, str):
                        record_field["type"] = field_type
                    else:
                        raise ValueError(f"Unexpected field type: {field_type}")
                    record_fields.append(record_field)
                avro_schema["fields"] = record_fields
            case EnumType(symbols=list(symbols)):
                avro_schema["type"] = "enum"
                avro_schema["symbols"] = symbols
            case ListType(values=values):
                avro_schema["type"] = "array"
                avro_schema["items"] = self._from_recap(values)
            case MapType(
                keys=StringType(),
                values=values,
            ):  # Avro only supports string keys
                avro_schema["type"] = "map"
                avro_schema["values"] = self._from_recap(values)
            case UnionType(types=list(types)):
                avro_schema["type"] = [self._from_recap(t) for t in types]
            case ProxyType(alias=str(alias)):
                if "." not in alias:
                    # Builtin types should be resolved
                    alias_type = self._from_recap(recap_type.resolve())
                    if isinstance(alias_type, dict):
                        avro_schema |= alias_type
                    else:
                        avro_schema["type"] = alias
                else:
                    # Fully qualified (non-builtin) aliases should just be referenced.
                    avro_schema["type"] = alias
                # Never redefine an alias. Needed because we set alias at the top of
                # the func, but the ProxyType alias param has a different semantic
                # meaning from other types.
                avro_schema.pop("aliases", None)
            case _:
                raise ValueError(f"Unsupported Recap type: {recap_type}")

        if len(avro_schema) == 1 and isinstance(avro_schema["type"], str):
            # Convert {"type": "type"} to "type"
            avro_schema = avro_schema["type"]

        return avro_schema

    def _from_recap_logical(self, recap_type: RecapType) -> dict[str, Any]:
        match recap_type:
            case BytesType(
                bytes_=int(bytes_),
                variable=bool(),
                logical=str(logical),
            ) if (
                bytes_ <= 9_223_372_036_854_775_807
                and logical == "build.recap.Decimal"
                and isinstance(recap_type.extra_attrs.get("precision"), int)
            ):
                return {
                    "type": "bytes",
                    "logicalType": "decimal",
                    "precision": recap_type.extra_attrs["precision"],
                    "scale": recap_type.extra_attrs.get("scale", 0),
                }
            case IntType(bits=int(bits), signed=True, logical=str(logical)) if (
                bits <= 32 and logical == "build.recap.Date"
            ):
                return {
                    "type": "int",
                    "logicalType": "date",
                }
            case IntType(bits=int(bits), signed=True, logical=str(logical)) if (
                bits <= 32
                and logical == "build.recap.Time"
                and recap_type.extra_attrs.get("unit") == "millisecond"
            ):
                return {
                    "type": "int",
                    "logicalType": "time-millis",
                }
            case IntType(bits=int(bits), signed=True, logical=str(logical)) if (
                bits <= 64
                and logical == "build.recap.Time"
                and recap_type.extra_attrs.get("unit") == "microsecond"
            ):
                return {
                    "type": "long",
                    "logicalType": "time-micros",
                }
            case IntType(bits=int(bits), signed=True, logical=str(logical)) if (
                bits <= 64
                and logical == "build.recap.Timestamp"
                and recap_type.extra_attrs.get("unit") == "millisecond"
            ):
                return {
                    "type": "long",
                    "logicalType": "timestamp-millis",
                }
            case IntType(bits=int(bits), signed=True, logical=str(logical)) if (
                bits <= 64
                and logical == "build.recap.Timestamp"
                and recap_type.extra_attrs.get("unit") == "microsecond"
            ):
                return {
                    "type": "long",
                    "logicalType": "timestamp-micros",
                }
            case BytesType(
                bytes_=12,
                variable=False,
                logical=str(logical),
            ) if (
                logical == "build.recap.Interval"
                and recap_type.extra_attrs.get("unit") == "millisecond"
            ):
                return {
                    "type": "fixed",
                    "logicalType": "duration",
                    "size": 12,
                }
            case StringType(bytes_=int(bytes_), logical=str(logical)) if (
                bytes_ <= 9_223_372_036_854_775_807 and logical == "build.recap.UUID"
            ):
                return {
                    "type": "string",
                    "logicalType": "uuid",
                }
            case _:
                raise ValueError(f"Unsupported Recap logical type: {recap_type}")

    def to_recap(
        self,
        avro_schema_str: str,
        namespace: str | None = None,
    ) -> StructType:
        avro_schema = json.loads(avro_schema_str)
        recap_schema = self._parse(avro_schema, namespace)
        if not isinstance(recap_schema, StructType):
            raise ValueError("Avro schema must be a record")
        return recap_schema

    def _parse(
        self,
        avro_schema: dict | str,
        namespace: str | None,
    ) -> RecapType:
        return_type = None
        alias: str | None = None
        extra_attrs = {}

        if isinstance(avro_schema, str):
            avro_schema = {"type": avro_schema}
        if "doc" in avro_schema:
            extra_attrs["doc"] = avro_schema["doc"]
        if "namespace" in avro_schema:
            namespace = avro_schema["namespace"]
        if "name" in avro_schema:
            name = avro_schema["name"]
            if "." in avro_schema["name"]:
                namespace, name = avro_schema["name"].rsplit(".", 1)
            if not namespace:
                namespace = "_root"
            extra_attrs["name"] = name
            alias = f"{namespace}.{name}"
        if "default" in avro_schema:
            extra_attrs["default"] = avro_schema["default"]
        if logical_type := avro_schema.get("logicalType"):
            return self._parse_logical(logical_type, avro_schema, extra_attrs)

        match avro_schema:
            case {"type": "null"}:
                return_type = NullType(alias=alias, **extra_attrs)
            case {"type": "boolean"}:
                return_type = BoolType(alias=alias, **extra_attrs)
            case {"type": "int"}:
                return_type = IntType(32, signed=True, alias=alias, **extra_attrs)
            case {"type": "long"}:
                return_type = IntType(64, signed=True, alias=alias, **extra_attrs)
            case {"type": "float"}:
                return_type = FloatType(32, alias=alias, **extra_attrs)
            case {"type": "double"}:
                return_type = FloatType(64, alias=alias, **extra_attrs)
            case {"type": "bytes"}:
                return_type = BytesType(
                    9_223_372_036_854_775_807,
                    alias=alias,
                    **extra_attrs,
                )
            case {
                "type": "fixed",
                "size": int(size),
            } if size <= 9_223_372_036_854_775_807:
                return_type = BytesType(
                    size,
                    alias=alias,
                    **extra_attrs,
                )
            case {"type": "string"}:
                return_type = StringType(
                    9_223_372_036_854_775_807,
                    alias=alias,
                    **extra_attrs,
                )
            case {"type": "record", "fields": fields}:
                fields = [
                    self._parse(field, namespace)
                    for field in avro_schema.get("fields", [])
                ]
                return_type = StructType(fields, alias=alias, **extra_attrs)
            case {"type": "enum", "symbols": symbols}:
                return_type = EnumType(symbols, alias=alias, **extra_attrs)
            case {"type": "array", "items": items}:
                values = self._parse(items, namespace)
                return_type = ListType(values, alias=alias, **extra_attrs)
            case {"type": "map", "values": values}:
                keys = StringType(9_223_372_036_854_775_807)
                return_type = MapType(
                    keys,
                    self._parse(values, namespace),
                    alias=alias,
                    **extra_attrs,
                )
            case {"type": "union", "types": types} | {"type": list(types)}:
                return_type = UnionType(
                    [self._parse(t, namespace) for t in types],
                    alias=alias,
                    **extra_attrs,
                )
            case {"type": str(type_alias)}:
                if "." not in type_alias:
                    type_alias = f"{namespace}.{type_alias}"

                # Short circuit so we don't re-register aliases for proxies
                return ProxyType(type_alias, self.registry, **extra_attrs)
            case {"type": dict(type_)}:
                return_type = self._parse(type_, namespace)

                # Don't pass extra_attrs into _parse since we don't want field
                # "name" to overwrite "alias" for record/enum/fixed.
                return_type.extra_attrs |= extra_attrs

                # Short circuit so we don't re-register aliases for nested types
                return return_type
            case _:
                raise ValueError(f"Unsupported Avro schema: {avro_schema}")

        if return_type.alias is not None:
            self.registry.register_alias(return_type)

        return return_type

    def _parse_logical(
        self,
        logical_type: str,
        avro_schema: dict,
        extra_attrs: dict,
    ) -> RecapType:
        match logical_type:
            case "decimal":
                return BytesType(
                    logical="build.recap.Decimal",
                    bytes_=avro_schema.get("size", 9_223_372_036_854_775_807),
                    variable=avro_schema["type"] == "fixed",
                    precision=avro_schema["precision"],
                    scale=avro_schema.get("scale", 0),
                    **extra_attrs,
                )
            case "uuid":
                return StringType(
                    logical="build.recap.UUID",
                    bytes_=36,
                    variable=False,
                    **extra_attrs,
                )
            case "date":
                return IntType(
                    logical="build.recap.Date",
                    bits=32,
                    signed=True,
                    unit="day",
                    **extra_attrs,
                )
            case "time-millis":
                return IntType(
                    logical="build.recap.Time",
                    bits=32,
                    signed=True,
                    unit="millisecond",
                    **extra_attrs,
                )
            case "time-micros":
                return IntType(
                    logical="build.recap.Time",
                    bits=64,
                    signed=True,
                    unit="microsecond",
                    **extra_attrs,
                )
            case "timestamp-millis":
                return IntType(
                    logical="build.recap.Timestamp",
                    bits=64,
                    signed=True,
                    unit="millisecond",
                    **extra_attrs,
                )
            case "timestamp-micros":
                return IntType(
                    logical="build.recap.Timestamp",
                    bits=64,
                    signed=True,
                    unit="microsecond",
                    **extra_attrs,
                )
            case "duration":
                return BytesType(
                    logical="build.recap.Interval",
                    bytes_=12,
                    signed=True,
                    unit="millisecond",
                    **extra_attrs,
                )
            case _:
                raise ValueError(f"Unsupported Avro logical type: {logical_type}")



import json

from fastapi import APIRouter, Depends, HTTPException, Request

from recap.settings import RecapSettings
from recap.storage.registry import RegistryStorage
from recap.types import RecapType, from_dict, to_dict

router = APIRouter(prefix="/registry")
settings = RecapSettings()


def get_storage() -> RegistryStorage:
    return RegistryStorage(
        settings.registry_storage_url.unicode_string(),
        **settings.registry_storage_url_args,
    )


@router.get("/")
async def ls(storage: RegistryStorage = Depends(get_storage)) -> list[str]:
    return storage.ls()


@router.get("/{name:str}")
async def latest(
    name: str,
    storage: RegistryStorage = Depends(get_storage),
) -> tuple[dict | list | str, int]:
    if type_and_version := storage.get(name):
        type_, version = type_and_version
        return to_dict(type_), version
    else:
        raise HTTPException(status_code=404, detail="Not found")


@router.get("/{name:str}/versions")
async def versions(
    name: str,
    storage: RegistryStorage = Depends(get_storage),
) -> list[int]:
    if versions := storage.versions(name):
        return versions
    else:
        raise HTTPException(status_code=404, detail="Not found")


@router.get("/{name:str}/versions/{version:int}")
async def version(
    name: str,
    version: int,
    storage: RegistryStorage = Depends(get_storage),
) -> tuple[dict | list | str, int]:
    if type_and_version := storage.get(name, version):
        type_, version = type_and_version
        return to_dict(type_), version
    else:
        raise HTTPException(status_code=404, detail="Not found")


@router.post("/{name:str}")
async def post(
    name: str,
    request: Request,
    storage: RegistryStorage = Depends(get_storage),
) -> int:
    type_ = await _request_to_type(request)

    return storage.put(name, type_)


@router.put("/{name:str}/versions/{version:int}")
async def put(
    name: str,
    version: int,
    request: Request,
    storage: RegistryStorage = Depends(get_storage),
):
    if storage.get(name, version):
        raise HTTPException(
            status_code=409,
            detail=f"Type {name} version {version} already exists",
        )

    type_ = await _request_to_type(request)

    return storage.put(name, type_, version)


async def _request_to_type(request: Request) -> RecapType:
    content_type = request.headers.get("content-type") or "application/x-recap+json"

    if content_type != "application/x-recap+json":
        raise HTTPException(
            status_code=415,
            detail=f"Unsupported content type: {content_type}",
        )

    type_bytes = await request.body()
    type_str = type_bytes.decode("utf-8")

    try:
        type_json = json.loads(type_str)
    except json.JSONDecodeError:
        # Assume payload is string alias if we can't decode as JSON.
        # If it's not, from_dict will fail below.
        type_json = type_str

    try:
        return from_dict(type_json)
    except Exception as e:
        raise HTTPException(
            status_code=400,
            detail=f"Failed to parse type: {e}",
        )



import pytest

from recap.storage.registry import RegistryStorage
from recap.types import IntType, StructType


@pytest.fixture
def registry_storage(tmp_path):
    storage_url = f"file://{tmp_path}"
    return RegistryStorage(storage_url)


def test_ls_empty(registry_storage):
    assert registry_storage.ls() == []


def test_put_get_single_item(registry_storage):
    name = "test"
    type_ = StructType(fields=[IntType(bits=24)])
    version = registry_storage.put(name, type_)

    retrieved_type, retrieved_version = registry_storage.get(name)

    assert retrieved_version == version
    assert retrieved_type == type_


def test_put_automatic_versioning(registry_storage):
    name = "test"
    type1 = StructType(fields=[IntType(bits=32)])
    type2 = StructType(fields=[IntType(bits=24)])

    version1 = registry_storage.put(name, type1)
    version2 = registry_storage.put(name, type2)

    assert version2 == version1 + 1


def test_versions(registry_storage):
    name = "test_versions"
    type_ = StructType(fields=[IntType(bits=32)])

    versions = [registry_storage.put(name, type_) for _ in range(5)]

    assert registry_storage.versions(name) == versions


def test_latest(registry_storage):
    name = "test_latest"
    type_ = StructType(fields=[IntType(bits=32)])

    versions = [registry_storage.put(name, type_) for _ in range(3)]
    assert registry_storage.latest(name) == max(versions)


def test_get_non_existent(registry_storage):
    assert registry_storage.get("non_existent") is None


def test_versions_non_existent(registry_storage):
    assert registry_storage.versions("non_existent") is None


def test_get_specific_version(registry_storage):
    name = "test_version"
    type1 = StructType(fields=[IntType(bits=48)])
    type2 = StructType(fields=[IntType(bits=24)])

    version1 = registry_storage.put(name, type1)
    version2 = registry_storage.put(name, type2)

    retrieved_type1, retrieved_version1 = registry_storage.get(name, version1)
    retrieved_type2, retrieved_version2 = registry_storage.get(name, version2)

    assert retrieved_version1 == version1
    assert retrieved_version2 == version2
    assert version1 + 1 == version2

    assert retrieved_type1 == type1
    assert retrieved_type2 == type2



from typing import Any

from recap.converters.dbapi import DbapiConverter
from recap.types import BoolType, BytesType, FloatType, IntType, RecapType, StringType


class SnowflakeConverter(DbapiConverter):
    def _parse_type(self, column_props: dict[str, Any]) -> RecapType:
        data_type = column_props["DATA_TYPE"].lower()
        octet_length = column_props["CHARACTER_OCTET_LENGTH"]

        if data_type in [
            "float",
            "float4",
            "float8",
            "double",
            "double precision",
            "real",
        ]:
            base_type = FloatType(bits=64)
        elif data_type == "boolean":
            base_type = BoolType()
        elif data_type in [
            "number",
            "decimal",
            "numeric",
            "int",
            "integer",
            "bigint",
            "smallint",
            "tinyint",
            "byteint",
        ] or (
            data_type.startswith("number")
            or data_type.startswith("decimal")
            or data_type.startswith("numeric")
        ):
            base_type = BytesType(
                logical="build.recap.Decimal",
                bytes_=16,
                variable=False,
                precision=column_props["NUMERIC_PRECISION"] or 38,
                scale=column_props["NUMERIC_SCALE"] or 0,
            )
        elif (
            data_type.startswith("varchar")
            or data_type.startswith("string")
            or data_type.startswith("text")
            or data_type.startswith("nvarchar")
            or data_type.startswith("nvarchar2")
            or data_type.startswith("char varying")
            or data_type.startswith("nchar varying")
        ):
            base_type = StringType(bytes_=octet_length or 16_777_216, variable=True)
        elif (
            data_type.startswith("char")
            or data_type.startswith("nchar")
            or data_type.startswith("character")
        ):
            base_type = StringType(bytes_=octet_length or 1, variable=True)
        elif data_type in ["binary", "varbinary", "blob"]:
            base_type = BytesType(bytes_=octet_length or 8_388_608)
        elif data_type == "date":
            base_type = IntType(bits=32, logical="build.recap.Date", unit="day")
        elif data_type.startswith("timestamp") or data_type.startswith("datetime"):
            params = self._parse_parameters(data_type)
            unit = self._get_time_unit(params) or "nanosecond"
            base_type = IntType(
                bits=64,
                logical="build.recap.Timestamp",
                unit=unit,
            )
        elif data_type.startswith("time"):
            params = self._parse_parameters(data_type)
            unit = self._get_time_unit(params) or "nanosecond"
            base_type = IntType(bits=32, logical="build.recap.Time", unit=unit)
        else:
            raise ValueError(f"Unknown data type: {data_type}")

        return base_type



import pytest

from recap.clients.bigquery import BigQueryClient


@pytest.mark.parametrize(
    "method, paths, expected_result",
    [
        # Test ls method
        ("ls", ["project1"], ("bigquery://", ["project1", None, None])),
        ("ls", [], ("bigquery://", [None, None, None])),
        (
            "ls",
            ["project1", "dataset1"],
            ("bigquery://", ["project1", "dataset1", None]),
        ),
        (
            "ls",
            ["project1", "dataset1", "table1"],
            ("bigquery://", ["project1", "dataset1", "table1"]),
        ),
        # Test schema method
        ("schema", ["project2"], ("bigquery://", ["project2", None, None])),
        (
            "schema",
            ["project2", "dataset2"],
            ("bigquery://", ["project2", "dataset2", None]),
        ),
        (
            "schema",
            ["project2", "dataset2", "table2"],
            ("bigquery://", ["project2", "dataset2", "table2"]),
        ),
        # Test invalid method
        (
            "invalid_method",
            ["project3"],
            pytest.raises(ValueError, match="Invalid method"),
        ),
    ],
)
def test_parse_method(method, paths, expected_result):
    if isinstance(expected_result, tuple):
        result = BigQueryClient.parse(method, paths)
        assert result == expected_result
    else:
        with expected_result:
            BigQueryClient.parse(method, paths)



from unittest.mock import MagicMock, patch

import fakesnow
import pytest
import snowflake.connector

from recap.clients.snowflake import SnowflakeClient
from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    NullType,
    StringType,
    StructType,
    UnionType,
)


class TestSnowflakeClient:
    @classmethod
    def setup_class(cls):
        with fakesnow.patch():
            cls.connection = snowflake.connector.connect()
            cursor = cls.connection.cursor()
            cursor.execute("CREATE OR REPLACE DATABASE testdb;")
            cursor.execute("USE DATABASE testdb;")
            cursor.execute("CREATE OR REPLACE SCHEMA public;")
            cursor.execute("USE SCHEMA testdb.public;")
            cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS test_types (
                    TEST_BIGINT BIGINT,
                    TEST_INTEGER INTEGER,
                    TEST_SMALLINT SMALLINT,
                    TEST_FLOAT FLOAT,
                    TEST_FLOAT8 FLOAT8,
                    TEST_FLOAT4 FLOAT4,
                    TEST_DOUBLE FLOAT,
                    TEST_DOUBLE_PRECISION DOUBLE PRECISION,
                    TEST_REAL REAL,
                    TEST_BOOLEAN BOOLEAN,
                    TEST_VARCHAR VARCHAR(100),
                    TEST_STRING STRING,
                    TEST_TEXT TEXT,
                    TEST_CHAR CHAR(10),
                    TEST_NVARCHAR NVARCHAR(100),
                    TEST_NVARCHAR2 NVARCHAR(100),
                    TEST_CHAR_VARYING CHAR VARYING(100),
                    TEST_NCHAR_VARYING NCHAR VARYING(100),
                    TEST_NCHAR NCHAR(100),
                    TEST_CHARACTER CHARACTER(100),
                    TEST_BINARY BINARY,
                    TEST_VARBINARY VARBINARY,
                    TEST_BLOB BLOB,
                    TEST_DATE DATE,
                    TEST_TIMESTAMP TIMESTAMP,
                    TEST_DATETIME DATETIME,
                    TEST_TIME TIME,
                    TEST_DECIMAL DECIMAL(10,2),
                    TEST_NUMERIC NUMERIC(10,2),
                    TEST_NUMBER NUMBER(10,2),
                    TEST_TINYINT TINYINT,
                    TEST_BYTEINT BYTEINT
                );
                """
            )

    def test_struct_method(self):
        client = SnowflakeClient(self.connection)  # type: ignore
        test_types_struct = client.schema("TESTDB", "PUBLIC", "TEST_TYPES")
        expected_fields = [
            UnionType(
                default=None,
                name="TEST_BIGINT",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=38,
                        scale=0,
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_INTEGER",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=38,
                        scale=0,
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_SMALLINT",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=38,
                        scale=0,
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_FLOAT",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="TEST_FLOAT8",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="TEST_FLOAT4",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="TEST_DOUBLE",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="TEST_DOUBLE_PRECISION",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="TEST_REAL",
                types=[NullType(), FloatType(bits=64)],
            ),
            UnionType(
                default=None,
                name="TEST_BOOLEAN",
                types=[NullType(), BoolType()],
            ),
            UnionType(
                default=None,
                name="TEST_VARCHAR",
                # 100 * 4 bytes per character = 400 bytes
                types=[NullType(), StringType(bytes_=400, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_STRING",
                types=[NullType(), StringType(bytes_=16_777_216, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_TEXT",
                types=[NullType(), StringType(bytes_=16_777_216, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_CHAR",
                types=[NullType(), StringType(bytes_=16_777_216, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_NVARCHAR",
                types=[NullType(), StringType(bytes_=16_777_216, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_NVARCHAR2",
                types=[NullType(), StringType(bytes_=16_777_216, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_CHAR_VARYING",
                # 100 * 4 bytes per character = 400 bytes
                types=[NullType(), StringType(bytes_=400, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_NCHAR_VARYING",
                # 100 * 4 bytes per character = 400 bytes
                types=[NullType(), StringType(bytes_=400, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_NCHAR",
                types=[NullType(), StringType(bytes_=16_777_216, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_CHARACTER",
                types=[NullType(), StringType(bytes_=16_777_216, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_BINARY",
                types=[NullType(), BytesType(bytes_=8_388_608, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_VARBINARY",
                types=[NullType(), BytesType(bytes_=8_388_608, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_BLOB",
                types=[NullType(), BytesType(bytes_=8_388_608, variable=True)],
            ),
            UnionType(
                default=None,
                name="TEST_DATE",
                types=[
                    NullType(),
                    IntType(bits=32, logical="build.recap.Date", unit="day"),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_TIMESTAMP",
                types=[
                    NullType(),
                    IntType(
                        bits=64,
                        logical="build.recap.Timestamp",
                        unit="nanosecond",
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_DATETIME",
                types=[
                    NullType(),
                    IntType(
                        bits=64,
                        logical="build.recap.Timestamp",
                        unit="nanosecond",
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_TIME",
                types=[
                    NullType(),
                    IntType(bits=32, logical="build.recap.Time", unit="nanosecond"),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_DECIMAL",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=10,
                        scale=2,
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_NUMERIC",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=10,
                        scale=2,
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_NUMBER",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=10,
                        scale=2,
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_TINYINT",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=38,
                        scale=0,
                    ),
                ],
            ),
            UnionType(
                default=None,
                name="TEST_BYTEINT",
                types=[
                    NullType(),
                    BytesType(
                        logical="build.recap.Decimal",
                        bytes_=16,
                        variable=False,
                        precision=38,
                        scale=0,
                    ),
                ],
            ),
        ]

        assert test_types_struct == StructType(fields=expected_fields)  # type: ignore

    def test_ls(self):
        client = SnowflakeClient(self.connection)  # type: ignore
        assert client.ls() == ["TESTDB"]
        assert client.ls("TESTDB") == [
            "PUBLIC",
            "information_schema",
            "main",
            "pg_catalog",
        ]
        assert client.ls("TESTDB", "PUBLIC") == ["TEST_TYPES"]

    def test_snowflake_client_create(self):
        with patch("snowflake.connector.connect") as mock_connect:
            mock_connection = MagicMock()
            mock_connect.return_value.__enter__.return_value = mock_connection

            with SnowflakeClient.create(
                host="some_account",
                paths=("some_database", "some_schema"),
            ) as _:
                mock_connect.assert_called_once_with(
                    account="some_account",
                    database="some_database",
                    schema="some_schema",
                    paths=("some_database", "some_schema"),
                )

    @pytest.mark.parametrize(
        "method, url_args, expected_result",
        [
            # Test "ls" method with only URL (no catalog or schema)
            (
                "ls",
                {"scheme": "snowflake", "netloc": "snowflake-url"},
                ("snowflake://snowflake-url", [None, None]),
            ),
            # Test "ls" method with URL and catalog
            (
                "ls",
                {"scheme": "snowflake", "netloc": "snowflake-url", "paths": ["cat1"]},
                ("snowflake://snowflake-url/cat1", ["cat1", None]),
            ),
            # Test "ls" method with URL, catalog, and schema
            (
                "ls",
                {
                    "scheme": "snowflake",
                    "netloc": "snowflake-url",
                    "paths": ["cat1", "sch1"],
                },
                ("snowflake://snowflake-url/cat1", ["cat1", "sch1"]),
            ),
            # Test "schema" method with URL, catalog, schema, and table
            (
                "schema",
                {
                    "scheme": "snowflake",
                    "netloc": "snowflake-url",
                    "paths": ["cat1", "sch1", "tbl1"],
                },
                ("snowflake://snowflake-url/cat1", ["cat1", "sch1", "tbl1"]),
            ),
            # Test invalid method
            (
                "invalid_method",
                {"scheme": "snowflake", "netloc": "snowflake-url"},
                pytest.raises(ValueError, match="Invalid method"),
            ),
            # Test "schema" method with insufficient paths (only catalog and schema, no table)
            (
                "schema",
                {
                    "scheme": "snowflake",
                    "netloc": "snowflake-url",
                    "paths": ["cat1", "sch1"],
                },
                pytest.raises(ValueError, match="Invalid method"),
            ),
            # Test "ls" method with all paths (catalog, schema, and table)
            (
                "ls",
                {
                    "scheme": "snowflake",
                    "netloc": "snowflake-url",
                    "paths": ["cat1", "sch1", "tbl1"],
                },
                pytest.raises(ValueError, match="Invalid method"),
            ),
        ],
    )
    def test_parse_method(self, method, url_args, expected_result):
        if isinstance(expected_result, tuple):
            result = SnowflakeClient.parse(method, **url_args)
            assert result == expected_result
        else:
            with expected_result:
                SnowflakeClient.parse(method, **url_args)
